"""Test FanController"""

import logging

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers import event
from aioafero.v1.controllers.fan import FanController, features

from .. import utils

zandra_fan = utils.create_devices_from_data("fan-ZandraFan.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.fans


@pytest.mark.asyncio
async def test_initialize_zandra(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "066c0e38-c49b-4f60-b805-486dc07cab74"
    assert dev.on == features.OnFeature(on=True)
    assert dev.speed == features.SpeedFeature(
        speed=50,
        speeds=[
            "fan-speed-6-016",
            "fan-speed-6-033",
            "fan-speed-6-050",
            "fan-speed-6-066",
            "fan-speed-6-083",
            "fan-speed-6-100",
        ],
    )
    assert dev.direction == features.DirectionFeature(forward=False)
    assert dev.device_information.model == "Zandra"


@pytest.mark.asyncio
async def test_turn_on(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = False
    await mocked_controller.turn_on(zandra_fan.id)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.on.on is True


@pytest.mark.asyncio
async def test_turn_off(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = True
    await mocked_controller.turn_off(zandra_fan.id)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.on.on is False


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "speed",
    [
        # Speed of 0 should turn off
        (0),
        # Exact value
        (16),
    ],
)
async def test_set_speed(speed, mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    if speed:
        dev.on.on = False
    else:
        dev.on.on = True
    dev.speed.speed = "fan-speed-6-100"
    await mocked_controller.set_speed(zandra_fan.id, speed)
    await mocked_controller._bridge.async_block_until_done()
    if not speed:
        assert dev.on.on is False
    else:
        assert dev.on.on is True
        assert dev.speed.speed == speed


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "on",
    [True, False],
)
@pytest.mark.parametrize(
    "forward",
    [
        (True),
        (False),
    ],
)
async def test_set_direction(on, forward, mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = on
    dev.direction.forward = not forward
    await mocked_controller.set_direction(zandra_fan.id, forward)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.direction.forward is forward


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "on",
    [
        (True,),
        (False,),
    ],
)
@pytest.mark.parametrize(
    "preset",
    [
        (True),
        (False),
    ],
)
async def test_set_preset(on, preset, mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = True
    dev.preset.enabled = not preset
    await mocked_controller.set_preset(zandra_fan.id, preset)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.preset.enabled is preset


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("fan-ZandraFan.json")[0]
    new_states = [
        AferoState(
            functionClass="toggle", value="disabled", lastUpdateTime=0, functionInstance="comfort-breeze"
        ),
        AferoState(
            functionClass="fan-speed", value="fan-speed-6-016", lastUpdateTime=0, functionInstance="fan-speed"
        ),
        AferoState(
            functionClass="fan-reverse", value="forward", lastUpdateTime=0, functionInstance="fan-reverse"
        ),
        AferoState(
            functionClass="power", value="off", lastUpdateTime=0, functionInstance="fan-power"
        ),
        AferoState(
            functionClass="toggle", value="disabled", lastUpdateTime=0, functionInstance="comfort-breeze"
        ),
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.preset.enabled is False
    assert dev.speed.speed == 16
    assert dev.direction.forward is True
    assert dev.on.on is False
    assert dev.available is False
    assert updates == {"speed", "direction", "on", "preset", "available"}


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    updates = await mocked_controller.update_elem(zandra_fan)
    assert updates == set()


# @TODO - Create tests for BaseResourcesController
@pytest.mark.asyncio
async def test_update(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(zandra_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.on.on is True
    manual_update = [
        {
            "functionClass": "power",
            "value": "off",
            "functionInstance": "fan-power",
        }
    ]
    mocked_controller._bridge.mock_update_afero_api(zandra_fan.id, manual_update)
    await mocked_controller.update(zandra_fan.id, states=manual_update)
    await mocked_controller._bridge.async_block_until_done()
    assert dev.on.on is False


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    await mocked_controller.set_state(zandra_fan.id)


@pytest.mark.asyncio
async def test_fan_emitting(bridge):
    dev_update = utils.create_devices_from_data("fan-ZandraFan.json")[0]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    await bridge.async_block_until_done()
    assert len(bridge.fans._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    await bridge.async_block_until_done()
    assert len(bridge.fans._items) == 1
    assert not bridge.fans._items[dev_update.id].available



"""Representation of an Afero Thermostat and its corresponding updates."""

from dataclasses import dataclass

from aioafero.util import calculate_hubspace_fahrenheit
from aioafero.v1.models import features

from .hvac_mixin import HVACMixin
from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class Thermostat(StandardMixin, HVACMixin):
    """Representation of an Afero Thermostat."""

    type: ResourceTypes = ResourceTypes.THERMOSTAT

    safety_max_temp: features.TargetTemperatureFeature | None = None
    safety_min_temp: features.TargetTemperatureFeature | None = None

    @property
    def target_temperature_range(self) -> tuple[float, float]:
        """Range which the thermostat supports."""
        if self.display_celsius:
            return (
                self.target_temperature_auto_heating.value,
                self.target_temperature_auto_cooling.value,
            )
        return (
            calculate_hubspace_fahrenheit(self.target_temperature_auto_heating.value),
            calculate_hubspace_fahrenheit(self.target_temperature_auto_cooling.value),
        )

    def get_mode_to_check(self) -> str | None:
        """Determine the current mode of the thermostat."""
        if not self.hvac_mode:
            return None
        if self.hvac_mode.mode in ["cool", "heat"]:
            return self.hvac_mode.mode
        if self.hvac_mode.previous_mode in ["cool", "heat"]:
            return self.hvac_mode.previous_mode
        return None

    @property
    def supports_fan_mode(self) -> bool:
        """Can enable fan-only mode."""
        return self.fan_mode is not None

    @property
    def supports_temperature_range(self) -> bool:
        """Range which the thermostat will heat / cool."""
        if not self.hvac_mode or "auto" not in self.hvac_mode.supported_modes:
            return False
        return (
            self.target_temperature_auto_cooling is not None
            and self.target_temperature_auto_heating is not None
        )


@dataclass
class ThermostatPut:
    """States that can be updated for a Thermostat."""

    fan_mode: features.ModeFeature | None = None
    hvac_mode: features.HVACModeFeature | None = None
    safety_max_temp: features.TargetTemperatureFeature | None = None
    safety_min_temp: features.TargetTemperatureFeature | None = None
    target_temperature_auto_heating: features.TargetTemperatureFeature | None = None
    target_temperature_auto_cooling: features.TargetTemperatureFeature | None = None
    target_temperature_heating: features.TargetTemperatureFeature | None = None
    target_temperature_cooling: features.TargetTemperatureFeature | None = None



import pytest

from aioafero.v1.models import SecuritySystemKeypad, features, DeviceInformation


@pytest.fixture
def populated_entity():
    return SecuritySystemKeypad(
        _id="entity-1",
        available=True,
        selects={
            ("volume", "buzzer-volume"): features.SelectFeature(
                selected="volume-04",
                selects={
                    "volume-00",
                    "volume-01",
                    "volume-02",
                    "volume-03",
                    "volume-04",
                },
                name="Buzzer Volume",
            ),
        },
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


def test_init(populated_entity):
    assert populated_entity.id == "entity-1"
    assert populated_entity.available is True
    assert populated_entity.instances == {"preset": "preset-1"}


def test_get_instance(populated_entity):
    assert populated_entity.get_instance("preset") == "preset-1"



"""Controller holding and managing Afero IoT resources of type `security-system`."""

from aioafero.device import AferoDevice
from aioafero.errors import DeviceNotFound
from aioafero.v1.models import SecuritySystemSensor, SecuritySystemSensorPut, features
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .base import AferoBinarySensor, AferoSensor, BaseResourcesController
from .security_system import (
    BYPASS_MODES,
    GENERIC_MODES,
    SENSOR_SPLIT_IDENTIFIER,
    TRIGGER_MODES,
)


class SecuritySystemSensorController(BaseResourcesController[SecuritySystemSensor]):
    """Controller holding and managing Afero IoT `security-system` for the sensors."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.SECURITY_SYSTEM_SENSOR]
    ITEM_CLS = SecuritySystemSensor
    ITEM_MAPPING = {}

    SENSOR_TYPES = {
        1: "Motion Sensor",
        2: "Door/Window Sensor",
    }

    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "tampered": "On",
        "triggered": "On",
    }
    # Elements that map to selects. func class / func instance to name
    ITEM_SELECTS = {
        ("chirpMode", None): "Chime",
        ("triggerType", None): "Alarming State",
        ("bypassType", None): "Bypass",
    }

    async def initialize_elem(self, device: AferoDevice):
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        available: bool = False
        numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = {}
        selects: dict[tuple[str, str | None], features.SelectFeature] | None = {}
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        device_type: int | None = None
        config_key: str | None = None
        for state in device.states:
            if select := await self.initialize_select(device.functions, state):
                selects[select[0]] = select[1]
            elif state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, device.id):
                # Security System Sensors only have binary sensors
                binary_sensors[sensor.id] = sensor
            elif state.functionClass == "top-level-key":
                config_key = state.value
        self._items[device.id] = SecuritySystemSensor(
            _id=device.id,
            split_identifier=SENSOR_SPLIT_IDENTIFIER,
            config_key=config_key,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            numbers=numbers,
            selects=selects,
            device_information=DeviceInformation(
                device_class=device.device_class,
                default_image=device.default_image,
                default_name=device.default_name,
                manufacturer=device.manufacturerName,
                model=self.SENSOR_TYPES.get(device_type, "Unknown"),
                name=device.friendly_name,
                parent_id=device.device_id,
                functions=device.functions,
            ),
        )
        return self._items[device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set[str]:
        """Update the Security System Sensor with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    updated_keys.add("available")
                cur_item.available = state.value
            elif (update_key := await self.update_sensor(state, cur_item)) or (
                update_key := await self.update_select(state, cur_item)
            ):
                updated_keys.add(update_key)
        return updated_keys

    async def set_state(
        self,
        device_id: str,
        selects: dict[tuple[str, int | None], str] | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = SecuritySystemSensorPut()
        try:
            cur_item: SecuritySystemSensor = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if selects:
            chirp_modes = {y: x for x, y in GENERIC_MODES.items()}
            trigger_types = {y: x for x, y in TRIGGER_MODES.items()}
            bypass_types = {y: x for x, y in BYPASS_MODES.items()}
            # Load the current values as it all needs to be sent
            select_vals = {
                "chirpMode": chirp_modes[
                    cur_item.selects.get(("chirpMode", None)).selected
                ],
                "triggerType": trigger_types[
                    cur_item.selects.get(("triggerType", None)).selected
                ],
                "bypassType": bypass_types[
                    cur_item.selects.get(("bypassType", None)).selected
                ],
            }
            for select, select_val in selects.items():
                if select[0] == "chirpMode":
                    select_vals["chirpMode"] = chirp_modes[select_val]
                elif select[0] == "triggerType":
                    select_vals["triggerType"] = trigger_types[select_val]
                elif select[0] == "bypassType":
                    select_vals["bypassType"] = bypass_types[select_val]
                else:
                    continue
            update_obj.sensor_config = features.SecuritySensorConfigFeature(
                sensor_id=cur_item.instance, key_name=cur_item.config_key, **select_vals
            )
            await self.update(cur_item.id, obj_in=update_obj)



from dataclasses import dataclass, field, fields, replace
import logging

import pytest
import asyncio
from aioafero import AferoDevice, AferoState
from aioafero.device import get_afero_device
from aioafero.errors import DeviceNotFound, ExceededMaximumRetries
from aioafero.v1 import AferoBridgeV1, models, v1_const
from aioafero.v1.controllers import event
from aioafero.v1.controllers.base import (
    ID_FILTER_ALL,
    BaseResourcesController,
    NumbersFeature,
    NumbersName,
    dataclass_to_afero,
    get_afero_instance_for_state,
    get_afero_state_from_feature,
    get_afero_states_from_list,
    get_afero_states_from_mapped,
)
from aioafero.v1.models.features import SelectFeature
from aioafero.v1.models.resource import DeviceInformation
import pytest_asyncio

from .. import utils


@dataclass
class TestFeatureBool:
    on: bool

    @property
    def api_value(self):
        return self.on


@dataclass
class TestFeatureInstance:
    on: bool
    func_instance: str | None

    @property
    def api_value(self):
        return {
            "value": "on" if self.on else "off",
            "functionClass": "mapped_beans",
            "functionInstance": self.func_instance or "",
        }


@dataclass
class ReturnsAListFeature:
    useless_value: bool

    @property
    def api_value(self):
        return [
            {
                "value": "cool",
                "functionClass": "bean-type",
                "functionInstance": "temperature",
            },
            {
                "value": "bean",
                "functionClass": "bean-type",
                "functionInstance": "warmth",
            },
        ]


@dataclass
class TestResource:
    id: str
    available: bool
    on: TestFeatureBool
    beans: dict[str | None, TestFeatureInstance]
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)


@dataclass
class TestResourceWithFunctions:
    id: str
    available: bool
    on: TestFeatureBool
    beans: dict[str | None, TestFeatureInstance]
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        instances = {}
        for function in functions:
            instances[function["functionClass"]] = function.get(
                "functionInstance", None
            )
        self.instances = instances

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)


test_res_funcs = TestResourceWithFunctions(
    [{"functionClass": "on", "functionInstance": "super-beans"}],
    id="cool",
    available=True,
    on=TestFeatureBool(on=True),
    beans={
        None: TestFeatureInstance(on=True, func_instance=None),
        "bean1": TestFeatureInstance(on=True, func_instance="bean1"),
        "bean2": TestFeatureInstance(on=False, func_instance="bean2"),
    },
    device_information=DeviceInformation(),
)


@dataclass
class TestResourcePut:
    on: TestFeatureBool | None
    beans: TestFeatureInstance | None


@dataclass
class TestResourcePutCallback:
    on: TestFeatureBool | None
    beans: TestFeatureInstance | None
    callback: callable


@dataclass
class TestResourceList:
    the_beans: ReturnsAListFeature | None


@dataclass
class TestResourceListPut:
    the_beans: ReturnsAListFeature | None = None


@dataclass
class TestResourceDict:
    id: str
    available: bool
    selects: dict[tuple[str, str | None], SelectFeature] | None
    device_information: DeviceInformation = field(default_factory=DeviceInformation)


@dataclass
class TestResourceDictPut:
    selects: dict[tuple[str, str | None], SelectFeature] | None


test_res = TestResource(
    id="cool",
    available=True,
    on=TestFeatureBool(on=True),
    beans={
        None: TestFeatureInstance(on=True, func_instance=None),
        "bean1": TestFeatureInstance(on=True, func_instance="bean1"),
        "bean2": TestFeatureInstance(on=False, func_instance="bean2"),
    },
    device_information=DeviceInformation(
        device_class="light",
    ),
)


test_res_update = TestResource(
    id="cool",
    available=True,
    on=TestFeatureBool(on=True),
    beans={
        None: TestFeatureInstance(on=True, func_instance=None),
        "bean1": TestFeatureInstance(on=True, func_instance="bean1"),
        "bean2": TestFeatureInstance(on=True, func_instance="bean2"),
    },
    device_information=DeviceInformation(
        device_class="light",
    ),
)

test_res_update_multiple = TestResource(
    id="cool",
    available=True,
    on=TestFeatureBool(on=False),
    beans={
        None: TestFeatureInstance(on=True, func_instance=None),
        "bean1": TestFeatureInstance(on=True, func_instance="bean1"),
        "bean2": TestFeatureInstance(on=True, func_instance="bean2"),
    },
    device_information=DeviceInformation(
        device_class="light",
    ),
)


test_res_default_dict = TestResourceDict(
    id="cool",
    available=True,
    selects={
        ("b1", None): SelectFeature(
            selected="True", selects={"True", "False"}, name="b1"
        ),
        ("b2", "one"): SelectFeature(
            selected="beans", selects={"beans", "cool"}, name="b2-1"
        ),
        ("b2", "two"): SelectFeature(
            selected="cool", selects={"beans", "cool"}, name="b2-2"
        ),
    },
    device_information=DeviceInformation(
        device_class="light",
    ),
)


test_res_update_dict = TestResourceDict(
    id="cool",
    available=True,
    selects={
        ("b1", None): SelectFeature(
            selected="False", selects={"True", "False"}, name="b1"
        ),
        ("b2", "one"): SelectFeature(
            selected="cool", selects={"beans", "cool"}, name="b2-1"
        ),
        ("b2", "two"): SelectFeature(
            selected="beans", selects={"beans", "cool"}, name="b2-2"
        ),
    },
    device_information=DeviceInformation(
        device_class="light",
    ),
)

test_device_dict = AferoDevice(
    id="cool",
    device_id="cool-parent",
    model="bean",
    device_class="light",
    default_name="bean",
    default_image="bean",
    friendly_name="bean",
    states=[
        AferoState(
            functionClass="on",
            value="on",
            lastUpdateTime=0,
            functionInstance=None,
        ),
        AferoState(
            functionClass="mapped_beans",
            value="on",
            lastUpdateTime=0,
        ),
        AferoState(
            functionClass="b1",
            value="True",
            lastUpdateTime=0,
            functionInstance=None,
        ),
        AferoState(
            functionClass="b2",
            value="beans",
            lastUpdateTime=0,
            functionInstance="one",
        ),
        AferoState(
            functionClass="b2",
            value="cool",
            lastUpdateTime=0,
            functionInstance="two",
        ),
    ],
    functions=[
        {
            "id": "af2f7826-990d-44bd-a8a1-d890438c7f1a",
            "functionClass": "b1",
            "functionInstance": None,
            "type": "category",
            "values": [
                {
                    "name": "True",
                },
                {
                    "name": "False",
                },
            ],
        },
        {
            "id": "af2f7826-990d-44bd-a8a1-d890438c7f1a",
            "functionClass": "b2",
            "functionInstance": "one",
            "type": "category",
            "values": [
                {
                    "name": "cool",
                },
                {
                    "name": "beans",
                },
            ],
        },
        {
            "id": "af2f7826-990d-44bd-a8a1-d890438c7f1a",
            "functionClass": "b2",
            "functionInstance": "two",
            "type": "category",
            "values": [
                {
                    "name": "cool",
                },
                {
                    "name": "beans",
                },
            ],
        },
    ],
)


test_device = AferoDevice(
    id="cool",
    device_id="cool-parent",
    model="bean",
    device_class="light",
    default_name="bean",
    default_image="bean",
    friendly_name="bean",
    states=[
        AferoState(
            functionClass="power",
            value="on",
            lastUpdateTime=0,
            functionInstance=None,
        ),
        AferoState(
            functionClass="mapped_beans",
            value="on",
            lastUpdateTime=0,
        ),
        AferoState(
            functionClass="mapped_beans",
            value="on",
            lastUpdateTime=0,
            functionInstance="bean1",
        ),
        AferoState(
            functionClass="mapped_beans",
            value="off",
            lastUpdateTime=0,
            functionInstance="bean2",
        ),
    ],
    functions=[],
)

test_device_update = AferoDevice(
    id="cool",
    device_id="cool-parent",
    model="bean",
    device_class="light",
    default_name="bean",
    default_image="bean",
    friendly_name="bean",
    states=[
        AferoState(
            functionClass="power",
            value="on",
            lastUpdateTime=0,
            functionInstance=None,
        ),
        AferoState(
            functionClass="mapped_beans",
            value="on",
            lastUpdateTime=0,
        ),
        AferoState(
            functionClass="mapped_beans",
            value="on",
            lastUpdateTime=0,
            functionInstance="bean1",
        ),
        AferoState(
            functionClass="mapped_beans",
            value="on",
            lastUpdateTime=0,
            functionInstance="bean2",
        ),
    ],
    functions=[],
)


def callback(polled_data: list[dict]):
    return event.CallbackResponse([], False)


class Example1ResourceController(BaseResourcesController):
    ITEM_TYPE_ID: models.ResourceTypes = models.ResourceTypes.DEVICE
    ITEM_TYPES: list[models.ResourceTypes] = [models.ResourceTypes.LIGHT]
    ITEM_CLS = TestResource
    ITEM_MAPPING: dict = {"beans": "mapped_beans"}
    DEVICE_SPLIT_CALLBACKS = {"nada": callback}
    ITEM_NUMBERS: dict[tuple[str, str | None], NumbersName] = {
        ("yup", None): NumbersName(unit="unit", display_name="disp"),
        ("nope", None): NumbersName(unit="unit", display_name=None),
        ("nope", "nope"): NumbersName(unit="unit", display_name=None),
    }

    async def initialize_elem(self, afero_dev: AferoDevice) -> TestResource:
        """Initialize the element"""
        self._logger.info("Initializing %s", afero_dev.id)
        on: TestFeatureBool | None = None
        beans: dict[str | None, TestFeatureInstance] = {}
        for state in afero_dev.states:
            if state.functionClass == "power":
                on = TestFeatureBool(on=state.value == "on")
            elif state.functionClass == "mapped_beans":
                beans[state.functionInstance] = TestFeatureInstance(
                    on=state.value == "on", func_instance=state.functionInstance
                )
        return TestResource(
            id=afero_dev.id,
            available=True,
            on=on,
            beans=beans,
            device_information=DeviceInformation(
                device_class=afero_dev.device_class,
            )
        )

    async def update_elem(self, afero_dev: AferoDevice) -> set:
        updated_keys = set()
        cur_item = self.get_device(afero_dev.id)
        for state in afero_dev.states:
            if state.functionClass == "power":
                new_val = state.value == "on"
                if cur_item.on.on != new_val:
                    updated_keys.add("on")
                    cur_item.on.on = new_val
            elif state.functionClass == "mapped_beans":
                new_val = state.value == "on"
                if cur_item.beans[state.functionInstance].on != new_val:
                    updated_keys.add("on")
                    cur_item.beans[state.functionInstance].on = state.value == "on"
            elif (update_key := await self.update_sensor(state, cur_item)) or (update_key := await self.update_number(state, cur_item)) or (update_key := await self.update_select(state, cur_item)):
                updated_keys.add(update_key)
        return updated_keys


class Example2ResourceController(BaseResourcesController):
    ITEM_TYPE_ID: models.ResourceTypes = models.ResourceTypes.DEVICE
    ITEM_TYPES: list[models.ResourceTypes] = [models.ResourceTypes.LIGHT]
    ITEM_CLS = TestResource
    ITEM_MAPPING: dict = {}
    ITEM_SELECTS = {
        ("b1", None): "b1",
        ("b2", "one"): "b2-1",
        ("b2", "two"): "b2-2",
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> TestResourceDict:
        """Initialize the element"""
        self._logger.info("Initializing %s", afero_device.id)
        selects: dict[tuple[str, str], SelectFeature] = {}
        for state in afero_device.states:
            if select := await self.initialize_select(afero_device.functions, state):
                selects[select[0]] = select[1]
        return TestResourceDict(
            id=afero_device.id,
            available=True,
            selects=selects,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
            )
        )

    async def update_elem(self, afero_dev: AferoDevice) -> set:
        updated_keys = set()
        cur_item = self.get_device(afero_dev.id)
        for state in afero_dev.states:
            if update_key := await self.update_select(state, cur_item):
                updated_keys.add(update_key)
        return updated_keys


@pytest_asyncio.fixture
async def ex1_rc(bridge_with_acct_req):
    bridge_with_acct_req.add_controller("ex1", Example1ResourceController)
    await bridge_with_acct_req.initialize()
    yield bridge_with_acct_req.ex1


@pytest_asyncio.fixture
async def ex2_rc(bridge_with_acct_req):
    bridge_with_acct_req.add_controller("ex2", Example2ResourceController)
    await bridge_with_acct_req.initialize()
    yield bridge_with_acct_req.ex2


def test_init(ex1_rc):
    assert isinstance(ex1_rc._bridge, AferoBridgeV1)
    assert ex1_rc._items == {}
    ex1_rc._initialized = False
    assert not ex1_rc.initialized
    ex1_rc._initialized = True
    assert ex1_rc.initialized
    assert ex1_rc.subscribers == {ID_FILTER_ALL: []}


def test_basic(ex1_rc):
    ex1_rc._items = {"cool": "beans"}
    assert ex1_rc["cool"] == "beans"
    assert "cool" in ex1_rc
    for item in ex1_rc:
        assert item == "beans"
    assert ex1_rc.items == ["beans"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "init_elem, evt_type, item_id, evt_data, expected_devs, expected_return",
    [
        # Device added
        (
            [],
            event.EventType.RESOURCE_ADDED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RESOURCE_ADDED,
                device_id=test_device.id,
                device=test_device,
                force_forward=False,
            ),
            {test_device.id},
            replace(test_res),
        ),
        # Device not found
        (
            [],
            event.EventType.RESOURCE_UPDATED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RESOURCE_UPDATED,
                device_id=test_device.id,
                device=test_device_update,
                force_forward=False,
            ),
            set(),
            None,
        ),
        # Device updated with changes
        (
            [test_device],
            event.EventType.RESOURCE_UPDATED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RESOURCE_UPDATED,
                device_id=test_device.id,
                device=test_device_update,
                force_forward=False,
            ),
            {test_device.id},
            replace(test_res_update),
        ),
        # Device updated with no changes + dont force
        (
            [test_device],
            event.EventType.RESOURCE_UPDATED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RESOURCE_UPDATED,
                device_id=test_device.id,
                device=test_device,
                force_forward=False,
            ),
            {test_device.id},
            None,
        ),
        # Device updated with no changes + force
        (
            [test_device],
            event.EventType.RESOURCE_UPDATED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RESOURCE_UPDATED,
                device_id=test_device.id,
                device=test_device,
                force_forward=True,
            ),
            {test_device.id},
            replace(test_res),
        ),
        # Device deleted
        (
            [test_device],
            event.EventType.RESOURCE_DELETED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RESOURCE_DELETED,
                device_id=test_device.id,
                force_forward=False,
            ),
            set(),
            replace(test_res),
        ),
        # Not a real event
        (
            [],
            event.EventType.RECONNECTED,
            test_device.id,
            event.AferoEvent(
                type=event.EventType.RECONNECTED,
                device_id=test_device.id,
                force_forward=False,
            ),
            set(),
            None,
        ),
    ],
)
async def test__handle_event_type(
    init_elem, evt_type, item_id, evt_data, expected_devs, expected_return, ex1_rc
):
    for elem in init_elem:
        ex1_rc._items[elem.id] = await ex1_rc.initialize_elem(elem)
        ex1_rc._bridge.add_device(elem.id, ex1_rc)
    assert (
        await ex1_rc._handle_event_type(evt_type, item_id, evt_data) == expected_return
    )
    if expected_return and evt_type != event.EventType.RESOURCE_DELETED:
        assert item_id in ex1_rc
    assert ex1_rc._bridge.tracked_devices == expected_devs


@pytest.mark.asyncio
@pytest.mark.parametrize("is_coroutine", [True, False])
@pytest.mark.parametrize(
    "id_filter, event_filter, event_type, expected",
    [
        ("beans", event.EventType.RESOURCE_ADDED, event.EventType.RESOURCE_ADDED, True),
        (
            "beans",
            event.EventType.RESOURCE_ADDED,
            event.EventType.RESOURCE_UPDATED,
            False,
        ),
        (
            "not-a-bean",
            event.EventType.RESOURCE_ADDED,
            event.EventType.RESOURCE_ADDED,
            False,
        ),
    ],
)
async def test_emit_to_subscribers(
    is_coroutine, id_filter, event_filter, event_type, expected, ex1_rc, mocker
):

    callback = mocker.AsyncMock() if is_coroutine else mocker.Mock()
    ex1_rc.subscribe(callback, id_filter=id_filter, event_filter=event_filter)
    await ex1_rc.emit_to_subscribers(event_type, "beans", test_res)
    if expected:
        callback.assert_called_once()
    else:
        callback.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "evt_type, evt_data, called",
    [
        # No data
        (event.EventType.RECONNECTED, None, False),
        # data
        (
            event.EventType.RESOURCE_DELETED,
            event.AferoEvent(
                type=event.EventType.RESOURCE_DELETED,
                device_id=test_device.id,
                force_forward=False,
            ),
            True,
        ),
        # data but not an item
        (
            event.EventType.RESOURCE_UPDATED,
            event.AferoEvent(
                type=event.EventType.RESOURCE_UPDATED,
                device_id="no-thanks",
                device=test_device,
                force_forward=True,
            ),
            False,
        ),
    ],
)
async def test__handle_event(evt_type, evt_data, called, ex1_rc, mocker):
    ex1_rc._items[test_device.id] = await ex1_rc.initialize_elem(test_device)
    ex1_rc._bridge.add_device(test_device.id, ex1_rc)
    emitted = mocker.patch.object(ex1_rc, "emit_to_subscribers")
    await ex1_rc._handle_event(evt_type, evt_data)
    if called:
        emitted.assert_called_once()
    else:
        emitted.assert_not_called()


def mocked_get_filtered_devices(initial_data) -> list[AferoDevice]:
    valid = []
    for ind, element in enumerate(initial_data):
        if element["typeId"] != models.ResourceTypes.DEVICE.value:
            continue
        if ind % 2 == 0:
            valid.append(get_afero_device(element))
    return valid


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "get_filtered_devices, expected_ids",
    [
        (
            False,
            {
                "99a03fb7-ebaa-4fc2-a7b5-df223003b127",
                "84338ebe-7ddf-4bfa-9753-3ee8cdcc8da6",
                "bc429efe-592a-4852-a18b-5b2a5e6ca5f1",
            },
        ),
        (
            mocked_get_filtered_devices,
            {
                "b16fc78d-4639-41a7-8a10-868405c412d6",
                "84338ebe-7ddf-4bfa-9753-3ee8cdcc8da6",
                "b50d9823-7ba0-44d9-b9a9-ad64dbbb225f",
            },
        ),
    ],
)
async def test__get_valid_devices(get_filtered_devices, expected_ids, ex1_rc):
    data = utils.get_raw_dump("raw_hs_data.json")
    if get_filtered_devices:
        ex1_rc.get_filtered_devices = get_filtered_devices
    devices = await ex1_rc._get_valid_devices(data)
    assert len(devices) == len(expected_ids)
    for device in devices:
        assert device.id in expected_ids


@pytest.mark.asyncio
async def test_initialize_not_needed(ex1_rc, mocker):
    check = mocker.patch.object(ex1_rc, "_get_valid_devices")
    ex1_rc._initialized = True
    await ex1_rc.initialize()
    check.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize("item_types", [True])
async def test_initialize(item_types, ex1_rc, mocker):
    ex1_rc._initialized = False
    handle_event = mocker.patch.object(ex1_rc, "_handle_event")
    await ex1_rc.initialize()
    assert len(ex1_rc._bridge.events._subscribers) == 14
    assert "nada" in ex1_rc._bridge.events.registered_multiple_devices
    assert ex1_rc._bridge.events.registered_multiple_devices["nada"] == callback


@pytest.mark.asyncio
@pytest.mark.parametrize(
    (
        "state",
        "expected_name",
        "expected_unit",
    ),
    [
        (
            AferoState(
                functionClass="yup",
                functionInstance=None,
                value=60,
            ),
            "disp",
            "unit",
        ),
        (
            AferoState(
                functionClass="nope",
                functionInstance=None,
                value=60,
            ),
            "nope",
            "unit",
        ),
        (
            AferoState(
                functionClass="nope",
                functionInstance="nope",
                value=60,
            ),
            "nope-nope",
            "unit",
        ),
    ],
)
async def test_initialize_number(state, expected_name, expected_unit, ex1_rc):
    func_def = {"values": [{"range": {"min": 60, "max": 1800, "step": 60}}]}
    _, number = await ex1_rc.initialize_number(func_def, state)
    assert number == NumbersFeature(
        value=state.value,
        min=60,
        max=1800,
        step=60,
        name=expected_name,
        unit=expected_unit,
    )


@pytest.mark.parametrize(
    "id_filter, event_filter, expected, expected_unsub",
    [
        # No ID filter
        (None, None, {"*": [(min, None)]}, {"*": []}),
        # ID filter
        ("beans", None, {"*": [], "beans": [(min, None)]}, {"*": [], "beans": []}),
        # ID filter as a tuple
        (
            ("beans", "double_beans"),
            (event.EventType.RESOURCE_ADDED,),
            {
                "*": [],
                "beans": [(min, (event.EventType.RESOURCE_ADDED,))],
                "double_beans": [(min, (event.EventType.RESOURCE_ADDED,))],
            },
            {"*": [], "beans": [], "double_beans": []},
        ),
    ],
)
def test_subscribe(id_filter, event_filter, expected, expected_unsub, ex1_rc):
    unsub = ex1_rc.subscribe(min, id_filter=id_filter, event_filter=event_filter)
    assert ex1_rc._subscribers == expected
    unsub()
    assert ex1_rc._subscribers == expected_unsub


def test_subscribe_with_starting(ex1_rc):
    ex1_rc.subscribe(min, id_filter="cool")
    unsub2 = ex1_rc.subscribe(min, id_filter="cool2")
    assert ex1_rc._subscribers == {
        "*": [],
        "cool": [(min, None)],
        "cool2": [(min, None)],
    }
    unsub2()
    assert ex1_rc._subscribers == {"*": [], "cool": [(min, None)], "cool2": []}


@pytest.mark.asyncio
async def test__process_state_update(ex1_rc):
    ex1_rc._items[test_res.id] = await ex1_rc.initialize_elem(test_device)
    ex1_rc._bridge.add_device(test_res.id, ex1_rc)
    await ex1_rc._process_state_update(
        ex1_rc._items[test_res.id],
        test_res.id,
        [
            {
                "functionClass": "mapped_beans",
                "value": "off",
                "lastUpdateTime": 0,
                "functionInstance": "bean2",
            }
        ],
    )
    assert ex1_rc._items[test_res.id].beans["bean2"].on is False
    update_req = await ex1_rc._bridge.events._event_queue.get()
    assert update_req["device"].id == "cool"
    assert update_req["device_id"] == "cool"
    assert update_req["force_forward"] is True
    assert update_req["type"] == event.EventType.RESOURCE_UPDATED
    state_update = update_req["device"].states[0]
    assert state_update.functionClass == "mapped_beans"
    assert state_update.functionInstance == "bean2"
    assert state_update.value == "off"


@pytest.mark.anyio
@pytest.mark.parametrize(
    "response, response_err, states, expected_call, expected, messages",
    [
        # Happy path
        (
            {"status": 200},
            None,
            [
                {
                    "functionClass": "mapped_beans",
                    "value": "off",
                    "lastUpdateTime": 0,
                    "functionInstance": "bean2",
                }
            ],
            {
                "json": {
                    "metadeviceId": "cool",
                    "values": [
                        {
                            "functionClass": "mapped_beans",
                            "value": "off",
                            "lastUpdateTime": 0,
                            "functionInstance": "bean2",
                        }
                    ],
                },
                "headers": {
                    "host": v1_const.AFERO_CLIENTS["hubspace"]["API_DATA_HOST"],
                    "content-type": "application/json; charset=utf-8",
                },
            },
            True,
            [],
        ),
        # Failed update
        (
            {"status": 400},
            None,
            [
                {
                    "functionClass": "mapped_beans",
                    "value": "off",
                    "lastUpdateTime": 0,
                    "functionInstance": "bean2",
                }
            ],
            {
                "json": {
                    "metadeviceId": "cool",
                    "values": [
                        {
                            "functionClass": "mapped_beans",
                            "value": "off",
                            "lastUpdateTime": 0,
                            "functionInstance": "bean2",
                        }
                    ],
                },
                "headers": {
                    "host": v1_const.AFERO_CLIENTS["hubspace"]["API_DATA_HOST"],
                    "content-type": "application/json; charset=utf-8",
                },
            },
            False,
            ["Invalid update provided for cool using"],
        ),
        # Retry exception
        (
            None,
            ExceededMaximumRetries,
            [
                {
                    "functionClass": "mapped_beans",
                    "value": "off",
                    "lastUpdateTime": 0,
                    "functionInstance": "bean2",
                }
            ],
            {
                "json": {
                    "metadeviceId": "cool",
                    "values": [
                        {
                            "functionClass": "mapped_beans",
                            "value": "off",
                            "lastUpdateTime": 0,
                            "functionInstance": "bean2",
                        }
                    ],
                },
                "headers": {
                    "host": v1_const.AFERO_CLIENTS["hubspace"]["API_DATA_HOST"],
                    "content-type": "application/json; charset=utf-8",
                },
            },
            False,
            ["Maximum retries exceeded for cool"],
        ),
    ],
)
async def test_update_afero_api(
    response,
    response_err,
    states,
    expected_call,
    expected,
    messages,
    ex1_rc,
    mock_aioresponse,
    caplog,
):
    device_id = "cool"
    url = ex1_rc._bridge.generate_api_url(v1_const.AFERO_GENERICS["API_DEVICE_STATE_ENDPOINT"].format(
        ex1_rc._bridge.account_id, str(device_id)
    ))
    if response:
        mock_aioresponse.put(url, **response)
    if response_err:
        ex1_rc._bridge.request.side_effect = response_err
    if expected:
        assert await ex1_rc.update_afero_api(device_id, states) is not False
    else:
        assert await ex1_rc.update_afero_api(device_id, states) == expected
    if expected_call:
        ex1_rc._bridge.request.assert_called_with("put", url, **expected_call)
    else:
        ex1_rc._bridge.request.assert_not_called()
    for message in messages:
        assert message in caplog.text


@pytest.mark.asyncio
async def test_update_dev_not_found(ex1_rc, caplog):
    caplog.set_level(logging.DEBUG)
    await ex1_rc.update("not-a-device")
    assert "Unable to update device not-a-device as it does not exist" in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "obj_in, states, expected_states, expected_item, successful",
    [
        # Obj in without updates
        (TestResourcePut(on=None, beans=None), None, None, test_res, True),
        # Obj in with updates
        (
            TestResourcePut(
                on=None, beans=TestFeatureInstance(on=True, func_instance="bean2")
            ),
            None,
            [
                {
                    "functionClass": "mapped_beans",
                    "functionInstance": "bean2",
                    "value": "on",
                    "lastUpdateTime": 12345,
                }
            ],
            test_res_update,
            True,
        ),
        # Obj in with unsuccessful updates
        (
            TestResourcePut(
                on=None, beans=TestFeatureInstance(on=True, func_instance="bean2")
            ),
            None,
            [
                {
                    "functionClass": "mapped_beans",
                    "functionInstance": "bean2",
                    "value": "on",
                    "lastUpdateTime": 12345,
                }
            ],
            test_res,
            False,
        ),
        # Manual states
        (
            None,
            [
                {
                    "functionClass": "mapped_beans",
                    "functionInstance": "bean2",
                    "value": "on",
                    "lastUpdateTime": 123456,
                }
            ],
            [
                {
                    "functionClass": "mapped_beans",
                    "functionInstance": "bean2",
                    "value": "on",
                    "lastUpdateTime": 123456,
                }
            ],
            test_res_update,
            True,
        ),
    ],
)
async def test_update(
    obj_in, states, expected_states, expected_item, successful, ex1_rc, mocker
):
    await ex1_rc.initialize()
    mocker.patch("time.time", return_value=12345)
    await ex1_rc._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(test_device)]
    )
    await ex1_rc._bridge.async_block_until_done()
    if successful:
        resp = mocker.AsyncMock()
        resp.status = 200
        json_resp = mocker.AsyncMock()
        json_resp.return_value = {"metadeviceId": test_device.id, "values": expected_states}
        resp.json = json_resp
        update_afero_api = mocker.patch.object(
            ex1_rc, "update_afero_api", return_value=resp
        )
    else:
        update_afero_api = mocker.patch.object(
            ex1_rc, "update_afero_api", return_value=False
        )
    await ex1_rc.update(test_device.id, obj_in=obj_in, states=states)
    if not expected_states:
        update_afero_api.assert_not_called()
    else:
        update_afero_api.assert_called_once_with(test_device.id, expected_states)
    await ex1_rc._bridge.async_block_until_done()
    assert ex1_rc._items[test_device.id] == expected_item


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "device, obj_in, states, expected_states, expected_item, successful",
    [
        (
            test_device_dict,
            TestResourceDictPut(
                selects={
                    ("b1", None): SelectFeature(
                        selected="False", selects={"True", "False"}, name="b1"
                    ),
                    ("b2", "one"): SelectFeature(
                        selected="cool", selects={"beans", "cool"}, name="b2-1"
                    ),
                    ("b2", "two"): SelectFeature(
                        selected="beans", selects={"beans", "cool"}, name="b2-2"
                    ),
                },
            ),
            None,
            [
                {
                    "functionClass": "b1",
                    "functionInstance": None,
                    "value": "False",
                    "lastUpdateTime": 12345,
                },
                {
                    "functionClass": "b2",
                    "functionInstance": "one",
                    "value": "cool",
                    "lastUpdateTime": 12345,
                },
                {
                    "functionClass": "b2",
                    "functionInstance": "two",
                    "value": "beans",
                    "lastUpdateTime": 12345,
                },
            ],
            test_res_update_dict,
            True,
        ),
    ],
)
async def test_update_dict(
    device, obj_in, states, expected_states, expected_item, successful, ex2_rc, mocker
):
    await ex2_rc.initialize()
    mocker.patch("time.time", return_value=12345)
    ex2_rc._items[device.id] = await ex2_rc.initialize_elem(device)
    ex2_rc._bridge.add_device(device.id, ex2_rc)
    ex2_rc._bridge.add_afero_dev(device)
    if successful:
        json_resp = mocker.AsyncMock(return_value={"metadeviceId": test_device.id, "values": expected_states})
    else:
        json_resp = mocker.AsyncMock(return_value=False)
    resp = mocker.AsyncMock()
    resp.json = json_resp
    update_afero_api = mocker.patch.object(
        ex2_rc, "update_afero_api", return_value=resp
    )
    await ex2_rc.update(device.id, obj_in=obj_in, states=states)
    if not expected_states:
        update_afero_api.assert_not_called()
    else:
        update_afero_api.assert_called_once_with(device.id, expected_states)
    await ex2_rc._bridge.async_block_until_done()
    assert ex2_rc._items[device.id] == expected_item


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "starting_items,device_id,expected",
    [
        ({"cool": "beans"}, "not-a-device", None),
        ({"cool": "beans"}, "cool", "beans"),
    ],
)
async def test_get_device(ex1_rc, starting_items, device_id, expected):
    ex1_rc._items = starting_items
    if not expected:
        with pytest.raises(DeviceNotFound):
            ex1_rc.get_device(device_id)
    else:
        assert ex1_rc.get_device(device_id) == expected


async def callback_test(elem, update_vals: dataclass):
    for f in fields(update_vals):
        if f.name == "callback":
            continue
        cur_val = getattr(update_vals, f.name, None)
        elem_val = getattr(elem, f.name)
        if cur_val is None:
            continue
        # There is probably a better way to approach this
        if not str(f.type).startswith("dict"):
            # Special processing for dicts
            if isinstance(elem_val, dict):
                cur_val = {getattr(cur_val, "func_instance", None): cur_val}
                getattr(elem, f.name).update(cur_val)
            else:
                setattr(elem, f.name, cur_val)
        else:
            elem_val.update(cur_val)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("elem", "update_obj", "mapping", "send_duplicate_states", "expected"),
    [
        # Empty
        (
            replace(test_res_default_dict),
            TestResourceDictPut(selects={}),
            {},
            False,
            [],
        ),
        # Dont match each other - test continue
        (
            replace(test_res_default_dict),
            TestResourcePut(on=None, beans=None),
            {},
            False,
            [],
        ),
        # No updates
        (
            replace(test_res_default_dict),
            TestResourceDictPut(
                selects={
                    ("b2", "two"): SelectFeature(
                        selected="cool", selects={"beans", "cool"}, name="b2-2"
                    ),
                }
            ),
            {},
            False,
            [],
        ),
        # Test dict mapping
        (
            replace(test_res_default_dict),
            TestResourceDictPut(
                selects={
                    ("b1", None): SelectFeature(
                        selected="False", selects={"True", "False"}, name="b1"
                    ),
                    ("b2", "one"): SelectFeature(
                        selected="cool", selects={"beans", "cool"}, name="b2-1"
                    ),
                    ("b2", "two"): SelectFeature(
                        selected="beans", selects={"beans", "cool"}, name="b2-2"
                    ),
                },
            ),
            {},
            False,
            [
                {
                    "functionClass": "b1",
                    "functionInstance": None,
                    "value": "False",
                    "lastUpdateTime": 12345,
                },
                {
                    "functionClass": "b2",
                    "functionInstance": "one",
                    "value": "cool",
                    "lastUpdateTime": 12345,
                },
                {
                    "functionClass": "b2",
                    "functionInstance": "two",
                    "value": "beans",
                    "lastUpdateTime": 12345,
                },
            ],
        ),
        # Test non-dict mapping
        (
            replace(test_res),
            TestResourcePut(
                on=False, beans=TestFeatureInstance(on=True, func_instance="bean2")
            ),
            {"on": "power"},
            False,
            [
                {
                    "functionClass": "power",
                    "functionInstance": None,
                    "lastUpdateTime": 12345,
                    "value": False,
                },
                {
                    "value": "on",
                    "functionClass": "mapped_beans",
                    "functionInstance": "bean2",
                    "lastUpdateTime": 12345,
                },
            ],
        ),
        # Testing a list
        (
            TestResourceList(the_beans=ReturnsAListFeature(useless_value=True)),
            TestResourceListPut(the_beans=ReturnsAListFeature(useless_value=False)),
            {},
            False,
            [
                {
                    "value": "cool",
                    "functionClass": "bean-type",
                    "functionInstance": "temperature",
                    "lastUpdateTime": 12345,
                },
                {
                    "value": "bean",
                    "functionClass": "bean-type",
                    "functionInstance": "warmth",
                    "lastUpdateTime": 12345,
                },
            ],
        ),
        # Testing when a value does not change
        (
            TestResourceList(the_beans=ReturnsAListFeature(useless_value=True)),
            TestResourceListPut(the_beans=ReturnsAListFeature(useless_value=True)),
            {},
False,
            [],
        ),
        # Test duplicates
        (
            replace(test_res_default_dict),
            TestResourceDictPut(
                selects={
                    ("b2", "two"): SelectFeature(
                        selected="cool", selects={"beans", "cool"}, name="b2-2"
                    ),
                }
            ),
            {},
            True,
             [
                 {
                     'functionClass': 'b2',
                     'functionInstance': 'two',
                     'lastUpdateTime': 12345,
                     'value': 'cool',
                    },
                ]
        ),
    ],
)
def test_dataclass_to_afero(elem, update_obj, mapping, send_duplicate_states, expected, mocker):
    mocker.patch("time.time", return_value=12345)
    assert dataclass_to_afero(elem, update_obj, mapping, send_duplicate_states) == expected


@pytest.mark.parametrize(
    ("element", "field_name", "update_vals", "send_duplicate_states", "expected"),
    [
        # Dont send dupes
        (
            replace(test_res_default_dict),
            "selects",
            {
                ("b1", None): SelectFeature(
                    selected="False", selects={"True", "False"}, name="b1"
                ),
                ("b2", "one"): SelectFeature(
                    selected="beans", selects={"beans", "cool"}, name="b2-1"
                ),
                ("b2", "two"): SelectFeature(
                    selected="beans", selects={"beans", "cool"}, name="b2-2"
                ),
            },
            False,
            [
                {
                    "functionClass": "b1",
                    "functionInstance": None,
                    "value": "False",
                    "lastUpdateTime": 12345,
                },
                {
                    "functionClass": "b2",
                    "functionInstance": "two",
                    "value": "beans",
                    "lastUpdateTime": 12345,
                },
            ],
        ),
        # Send dupes
(
            replace(test_res_default_dict),
            "selects",
            {
                ("b1", None): SelectFeature(
                    selected="False", selects={"True", "False"}, name="b1"
                ),
                ("b2", "one"): SelectFeature(
                    selected="beans", selects={"beans", "cool"}, name="b2-1"
                ),
                ("b2", "two"): SelectFeature(
                    selected="beans", selects={"beans", "cool"}, name="b2-2"
                ),
            },
            True,
            [
                {
                    "functionClass": "b1",
                    "functionInstance": None,
                    "value": "False",
                    "lastUpdateTime": 12345,
                },
{
                    "functionClass": "b2",
                    "functionInstance": "one",
                    "value": "beans",
                    "lastUpdateTime": 12345,
                },
                {
                    "functionClass": "b2",
                    "functionInstance": "two",
                    "value": "beans",
                    "lastUpdateTime": 12345,
                },
            ],
        )
    ],
)
def test_get_afero_states_from_mapped(
    element, field_name, update_vals, send_duplicate_states, expected, mocker
):
    mocker.patch("time.time", return_value=12345)
    assert get_afero_states_from_mapped(element, field_name, update_vals, send_duplicate_states) == expected


@pytest.mark.parametrize(
    "elem, feat, mapped_afero_key, expected",
    [
        # Utilize func_instance
        (test_res, TestFeatureInstance(on=True, func_instance="bean1"), None, "bean1"),
        # Utilize instances
        (test_res_funcs, TestFeatureBool(on=True), "on", "super-beans"),
        # None fallback
        (test_res_funcs, TestFeatureBool(on=True), None, None),
    ],
)
def test_get_afero_instance_for_state(elem, feat, mapped_afero_key, expected, mocker):
    mocker.patch("time.time", return_value=12345)
    assert get_afero_instance_for_state(elem, feat, mapped_afero_key) == expected


@pytest.mark.parametrize(
    "func_class, func_instance, current_val, expected",
    [
        # Non-dict val
        (
            "cool",
            "beans",
            "super-beans",
            {
                "functionClass": "cool",
                "functionInstance": "beans",
                "lastUpdateTime": 12345,
                "value": "super-beans",
            },
        ),
        # dict val
        (
            "cool",
            "beans",
            {"functionClass": "beans", "functionInstance": "cool", "value": 12},
            {
                "functionClass": "beans",
                "functionInstance": "cool",
                "lastUpdateTime": 12345,
                "value": 12,
            },
        ),
    ],
)
def test_get_afero_state_from_feature(
    func_class, func_instance, current_val, expected, mocker
):
    mocker.patch("time.time", return_value=12345)
    assert (
        get_afero_state_from_feature(func_class, func_instance, current_val) == expected
    )


@pytest.mark.parametrize(
    "states, expected",
    [
        (
            [
                {
                    "functionClass": "beans",
                    "functionInstance": "cool",
                    "value": 12,
                },
                {
                    "functionClass": "beans",
                    "functionInstance": "cool2",
                    "value": 12,
                },
            ],
            [
                {
                    "functionClass": "beans",
                    "functionInstance": "cool",
                    "lastUpdateTime": 12345,
                    "value": 12,
                },
                {
                    "functionClass": "beans",
                    "functionInstance": "cool2",
                    "lastUpdateTime": 12345,
                    "value": 12,
                },
            ],
        )
    ],
)
def test_get_afero_states_from_list(states, expected, mocker):
    mocker.patch("time.time", return_value=12345)
    assert get_afero_states_from_list(states) == expected



"""Representation of a top-level item."""

from dataclasses import dataclass

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class Device(StandardMixin):
    """Representation of an Afero parent item."""

    type: ResourceTypes = ResourceTypes.PARENT_DEVICE



"""Representation of an Afero Switch and its corresponding updates."""

from dataclasses import dataclass

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class Switch(StandardMixin):
    """Representation of an Afero Switch."""

    type: ResourceTypes = ResourceTypes.SWITCH

    on: dict[str | None, features.OnFeature] | None = None


@dataclass
class SwitchPut:
    """States that can be updated for a Switch."""

    on: features.OnFeature | None = None



"""Handle authentication to Afero API."""

__all__ = ["AferoAuth", "TokenData", "passthrough"]

import asyncio
import base64
from contextlib import contextmanager
import datetime
import hashlib
import logging
import os
import re
from typing import Final, NamedTuple
from urllib.parse import parse_qs, urlparse

import aiohttp
from aiohttp import ClientResponseError, ContentTypeError
from bs4 import BeautifulSoup
from securelogging import LogRedactorMessage, add_secret, remove_secret

from aioafero.errors import InvalidAuth, InvalidOTP, InvalidResponse, OTPRequired

from . import v1_const

logger = logging.getLogger(__name__)

TOKEN_TIMEOUT: Final[int] = 118
STATUS_CODE: Final[str] = "Status Code: %s"


class AuthChallenge(NamedTuple):
    """Data used to perform the initial authentication."""

    challenge: str
    verifier: str


class TokenData(NamedTuple):
    """Data related to the current token."""

    token: str
    access_token: str
    refresh_token: str
    expiration: float


class AuthSessionData(NamedTuple):
    """Data related to current attempt to login."""

    session_code: str
    execution: str
    tab_id: str


@contextmanager
def passthrough():
    """Do nothing."""
    yield


class AferoAuth:
    """Authentication against the Afero IoT API.

    This class follows the Afero IoT authentication workflow and utilizes
    refresh tokens.
    """

    def __init__(
        self,
        bridge,
        username,
        password,
        hide_secrets: bool = True,
        refresh_token: str | None = None,
        afero_client: str | None = "hubspace",
    ):
        """Create a class to handle authentication with Afero IoT API."""
        self.logger = logging.getLogger(f"{__package__}[{username}]")
        if hide_secrets:
            self.secret_logger = LogRedactorMessage
        else:
            self.secret_logger = passthrough
        self._hide_secrets: bool = hide_secrets
        self._async_lock: asyncio.Lock = asyncio.Lock()
        self._username: str = username
        self._password: str = password
        self._token_data: TokenData | None = None
        self._bridge = bridge
        if refresh_token:
            add_secret(refresh_token)
            self._token_data = TokenData(
                None, None, refresh_token, datetime.datetime.now().timestamp()
            )
        self._afero_client: str = afero_client
        self._token_headers: dict[str, str] = {
            "Content-Type": "application/x-www-form-urlencoded",
            "accept-encoding": "gzip",
            "host": v1_const.AFERO_CLIENTS[self._afero_client]["AUTH_OPENID_HOST"],
        }
        self._otp_data: dict = {}

    @property
    async def is_expired(self) -> bool:
        """Determine if the token is expired."""
        if not self._token_data:
            return True
        return datetime.datetime.now().timestamp() >= self._token_data.expiration

    @property
    def refresh_token(self) -> str | None:
        """Get the current refresh token."""
        if not self._token_data:
            return None
        return self._token_data.refresh_token

    def generate_auth_url(self, endpoint: str) -> str:
        """Generate an auth URL for the Afero API."""
        endpoint = endpoint.removeprefix("/")
        return f"https://{v1_const.AFERO_CLIENTS[self._afero_client]['AUTH_OPENID_HOST']}/auth/realms/{v1_const.AFERO_CLIENTS[self._afero_client]['AUTH_REALM']}/{endpoint}"

    def set_token_data(self, data: TokenData) -> None:
        """Set the current taken data."""
        self._token_data = data

    async def webapp_login(self, challenge: AuthChallenge) -> str:
        """Perform login to the webapp for a code.

        Login to the webapp and generate a code used for generating tokens.

        :param challenge: Challenge data for connection and approving

        :return: Code used for generating a refresh token
        """
        code_params: dict[str, str] = {
            "response_type": "code",
            "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                "AUTH_DEFAULT_CLIENT_ID"
            ],
            "redirect_uri": v1_const.AFERO_CLIENTS[self._afero_client][
                "AUTH_DEFAULT_REDIRECT_URI"
            ],
            "code_challenge": challenge.challenge,
            "code_challenge_method": "S256",
            "scope": "openid offline_access",
        }
        url = self.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_OPENID_ENDPOINT"])
        self.logger.debug(
            "URL: %s\n\tparams: %s",
            url,
            code_params,
        )
        try:
            response = await self._bridge.request(
                "GET",
                url,
                include_token=False,
                params=code_params,
                allow_redirects=False,
            )
        except aiohttp.web_exceptions.HTTPError as err:
            raise InvalidResponse("Unable to query login page") from err
        if response.status == 200:
            contents = await response.text()
            login_data = await extract_login_data(contents, "kc-form-login")
            self.logger.debug(
                ("WebApp Login:\n\tSession Code: %s\n\tExecution: %s\n\tTab ID:%s"),
                login_data.session_code,
                login_data.execution,
                login_data.tab_id,
            )
            return await self.generate_code(login_data, challenge)
        if response.status == 302:
            self.logger.debug("Hubspace returned an active session")
            return await AferoAuth.parse_code(response)
        try:
            response.raise_for_status()
        except ClientResponseError as err:
            raise InvalidResponse("Unable to query login page") from err

    @staticmethod
    async def generate_challenge_data() -> AuthChallenge:
        """Generate data to send to Afero API when logging into the system."""
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
        code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)
        code_challenge = hashlib.sha256(code_verifier.encode("utf-8")).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
        code_challenge = code_challenge.replace("=", "")
        chal = AuthChallenge(code_challenge, code_verifier)
        logger.debug("Challenge information: %s", chal)
        return chal

    async def generate_code(
        self, data: AuthSessionData, challenge: AuthChallenge
    ) -> str:
        """Finalize login to Afero IoT page.

        :param session_code: Session code during form interaction
        :param execution: Session code during form interaction
        :param tab_id: Session code during form interaction
        :return: code for generating tokens
        """
        self.logger.debug("Generating code")
        params = extract_login_codes(data, self._afero_client)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "x-requested-with": "io.afero.partner.hubspace",
        }
        auth_data = {
            "username": self._username,
            "password": self._password,
            "credentialId": "",
        }
        url = self.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_CODE_ENDPOINT"])
        self.logger.debug(
            "URL: %s\n\tparams: %s\n\theaders: %s",
            url,
            params,
            headers,
        )
        response = await self._bridge.request(
            "POST",
            url,
            include_token=False,
            params=params,
            data=auth_data,
            headers=headers,
            allow_redirects=False,
        )
        self.logger.debug(STATUS_CODE, response.status)
        # If OTP is required, a 200 will be returned. If a 302 is returned, then there
        # is no OTP required.
        content = await response.text()
        if await AferoAuth.requires_otp(content):
            login_data = await extract_login_data(content, "kc-otp-login-form")
            otp_params = extract_login_codes(login_data, self._afero_client)
            self._otp_data = {
                "params": otp_params,
                "headers": headers,
                "challenge": challenge,
            }
            raise OTPRequired
        if response.status != 302:
            raise InvalidAuth(
                "Unable to authenticate with the supplied username / password"
            )
        return await AferoAuth.parse_code(response)

    @staticmethod
    async def requires_otp(content: str) -> bool:
        """Determine if the user requires otp."""
        return "kc-otp-login-form" in content

    @staticmethod
    async def parse_code(response: aiohttp.ClientResponse) -> str:
        """Parse the code for generating tokens."""
        try:
            parsed_url = urlparse(response.headers["location"])
            code = parse_qs(parsed_url.query)["code"][0]
            logger.debug("Location: %s", response.headers.get("location"))
            logger.debug("Code: %s", code)
        except KeyError as err:
            raise InvalidResponse(
                f"Unable to process the result from {response.url}: {response.status}"
            ) from err
        return code

    async def generate_refresh_token(
        self,
        challenge: AuthChallenge | None = None,
        code: str | None = None,
    ) -> TokenData:
        """Generate a refresh token.

        If a challenge is provided, it will send the correct data. If no challenge is required,
        it will use the existing token

        :param client: async client for making request
        :param code: Code used for generating refresh token
        :param challenge: Challenge data for connection and approving

        :return: Refresh token to generate a new token
        """
        self.logger.debug("Generating refresh token")
        if challenge:
            data = {
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": v1_const.AFERO_CLIENTS[self._afero_client][
                    "AUTH_DEFAULT_REDIRECT_URI"
                ],
                "code_verifier": challenge.verifier,
                "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                    "AUTH_DEFAULT_CLIENT_ID"
                ],
            }
        else:
            data = {
                "grant_type": "refresh_token",
                "refresh_token": self._token_data.refresh_token,
                "scope": "openid email offline_access profile",
                "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                    "AUTH_DEFAULT_CLIENT_ID"
                ],
            }
        url = self.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_TOKEN_ENDPOINT"])
        with self.secret_logger():
            self.logger.debug(
                "URL: %s\n\tdata: %s\n\theaders: %s",
                url,
                data,
                self._token_headers,
            )
        response = await self._bridge.request(
            "POST",
            url,
            headers=self._token_headers,
            include_token=False,
            data=data,
        )
        self.logger.debug(STATUS_CODE, response.status)
        try:
            resp_json = await response.json()
        except (ValueError, ContentTypeError) as err:
            raise InvalidResponse(
                "Unexpected data returned during token refresh"
            ) from err
        if response.status != 200:
            if resp_json and resp_json.get("error") == "invalid_grant":
                raise InvalidAuth
            try:
                response.raise_for_status()
            except ClientResponseError as err:
                raise InvalidResponse(
                    "Unexpected data returned during token refresh"
                ) from err
        try:
            refresh_token = resp_json["refresh_token"]
            access_token = resp_json["access_token"]
            id_token = resp_json["id_token"]
        except KeyError as err:
            raise InvalidResponse("Unable to extract refresh token") from err
        add_secret(refresh_token)
        add_secret(access_token)
        add_secret(id_token)
        with self.secret_logger():
            self.logger.debug("JSON response: %s", resp_json)
        return TokenData(
            id_token,
            access_token,
            refresh_token,
            datetime.datetime.now().timestamp() + TOKEN_TIMEOUT,
        )

    async def perform_initial_login(self) -> TokenData:
        """Login to generate a refresh token.

        :return: Refresh token for the auth
        """
        challenge = await AferoAuth.generate_challenge_data()
        code: str = await self.webapp_login(challenge)
        self.logger.debug("Successfully generated an auth code")
        refresh_token = await self.generate_refresh_token(
            code=code, challenge=challenge
        )
        self.logger.debug("Successfully generated a refresh token")
        return refresh_token

    async def perform_otp_login(self, otp_code: str) -> TokenData:
        """Perform otp login to generate a refresh token.

        :return: Refresh token for the auth
        """
        if self._otp_data == {}:
            raise OTPRequired("No OTP data available to perform login")
        self.logger.debug("Performing otp login")
        code: str = await self.submit_otp(otp_code)
        refresh_token = await self.generate_refresh_token(
            code=code, challenge=self._otp_data["challenge"]
        )
        self.logger.debug("Successfully generated a refresh token")
        self._token_data = refresh_token
        self._otp_data = {}
        return refresh_token

    async def submit_otp(self, otp_code: str) -> None:
        """Submit OTP code to continue the login process.

        :param otp_code: OTP code provided by the user

        :return: Code used for generating a refresh token
        """
        otp_data = {
            "action": "submit",
            "flowName": "doLogIn",
            "emailCode": otp_code,
        }
        url = self.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_CODE_ENDPOINT"])
        response = await self._bridge.request(
            "POST",
            url,
            include_token=False,
            params=self._otp_data["params"],
            data=otp_data,
            headers=self._otp_data["headers"],
            allow_redirects=False,
        )
        if response.status != 302:
            self.logger.warning("OTP code was invalid. Re-enter the OTP code.")
            content = await response.text()
            kc_error = get_kc_error(content)
            login_data = await extract_login_data(content, "kc-otp-login-form")
            otp_params = extract_login_codes(login_data, self._afero_client)
            self._otp_data["params"] = otp_params
            raise InvalidOTP(kc_error)
        return await AferoAuth.parse_code(response)

    async def token(self, retry: bool = True) -> str:
        """Generate the token required to make Afero API calls."""
        invalidate_refresh_token = False
        async with self._async_lock:
            if not self._token_data:
                self.logger.debug(
                    "Refresh token not present. Generating a new refresh token"
                )
                self._token_data = await self.perform_initial_login()
            if await self.is_expired:
                self.logger.debug("Token has not been generated or is expired")
                try:
                    new_data = await self.generate_refresh_token()
                    remove_secret(self._token_data.token)
                    remove_secret(self._token_data.access_token)
                    remove_secret(self._token_data.refresh_token)
                    self._token_data = new_data
                except InvalidAuth:
                    self.logger.debug("Provided refresh token is no longer valid.")
                    if not retry:
                        raise
                    invalidate_refresh_token = True
                else:
                    self.logger.debug("Token has been successfully generated")
        if invalidate_refresh_token:
            return await self.token(retry=False)
        return self._token_data.token


async def extract_login_data(page: str, form_login_element: str) -> AuthSessionData:
    """Extract the required login data from the auth page.

    :param page: the response from performing a GET against
    v1_const.AFERO_CLIENTS[self._afero_client]['OPENID_URL']
    """
    auth_page = BeautifulSoup(page, features="html.parser")
    login_form = auth_page.find("form", id=form_login_element)
    if login_form is None:
        raise InvalidResponse("Unable to parse login page")
    try:
        login_url: str = login_form.attrs["action"]
    except KeyError as err:
        raise InvalidResponse("Unable to extract login url") from err
    parsed_url = urlparse(login_url)
    login_data = parse_qs(parsed_url.query)
    try:
        return AuthSessionData(
            login_data["session_code"][0],
            login_data["execution"][0],
            login_data["tab_id"][0],
        )
    except (KeyError, IndexError) as err:
        raise InvalidResponse("Unable to parse login url") from err


def extract_login_codes(data: AuthSessionData, client: str) -> dict:
    return {
        "session_code": data.session_code,
        "execution": data.execution,
        "client_id": v1_const.AFERO_CLIENTS[client]["AUTH_DEFAULT_CLIENT_ID"],
        "tab_id": data.tab_id,
    }


def get_kc_error(page: str) -> str:
    """Extract the error message from the otp page."""
    auth_page = BeautifulSoup(page, features="html.parser")
    error_div = auth_page.find("span", class_="kc-feedback-text")
    if error_div is None:
        return "Unknown error"
    return error_div.text.strip()



import pytest

from aioafero.v1.models import features, DeviceInformation
from aioafero.v1.models.fan import Fan


@pytest.fixture
def populated_fan():
    return Fan(
        _id="fan-1",
        available=True,
        on=features.OnFeature(on=True),
        speed=features.SpeedFeature(speed=50, speeds=[25, 50, 75, 100]),
        direction=features.DirectionFeature(forward=True),
        preset=features.PresetFeature(
            enabled=True, func_class="preset", func_instance="preset-1"
        ),
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


@pytest.fixture
def empty_fan():
    return Fan(
        _id="fan-1",
        available=True,
        on=None,
        speed=None,
        direction=None,
        preset=None,
    )


def test_init(populated_fan):
    assert populated_fan.id == "fan-1"
    assert populated_fan.available is True
    assert populated_fan.on.on is True
    assert populated_fan.speed.speed == 50
    assert populated_fan.direction.forward is True
    assert populated_fan.preset.enabled is True
    assert populated_fan.is_on is True
    populated_fan.on.on = False
    assert populated_fan.is_on is False
    assert populated_fan.current_direction is True
    assert populated_fan.current_speed == 50
    assert populated_fan.current_preset == "preset-1"
    populated_fan.preset.enabled = False
    assert populated_fan.current_preset is None


def test_empty_fan(empty_fan):
    assert not empty_fan.supports_presets
    assert not empty_fan.current_preset
    assert not empty_fan.supports_speed
    assert not empty_fan.current_speed
    assert not empty_fan.supports_direction
    assert not empty_fan.current_direction
    assert not empty_fan.supports_on
    assert not empty_fan.is_on



@pytest.mark.parametrize(
    "has_feature",
    [True, False],
)
def test_supports_direction(has_feature, populated_fan):
    if not has_feature:
        populated_fan.direction = None
    assert populated_fan.supports_direction == has_feature


@pytest.mark.parametrize(
    "has_feature",
    [True, False],
)
def test_supports_on(has_feature, populated_fan):
    if not has_feature:
        populated_fan.on = None
    assert populated_fan.supports_on == has_feature


@pytest.mark.parametrize(
    "has_feature",
    [True, False],
)
def test_supports_presets(has_feature, populated_fan):
    if not has_feature:
        populated_fan.preset = None
    assert populated_fan.supports_presets == has_feature


@pytest.mark.parametrize(
    "has_feature",
    [True, False],
)
def test_supports_speed(has_feature, populated_fan):
    if not has_feature:
        populated_fan.speed = None
    assert populated_fan.supports_speed == has_feature



"""Devices and States for API responses."""

__all__ = [
    "AferoDevice",
    "AferoResource",
    "AferoState",
    "get_afero_device",
    "get_function_from_device",
]
from dataclasses import dataclass, field, fields
import logging
from typing import Any, TypeVar

logger = logging.getLogger(__name__)


@dataclass
class AferoState:
    """State of a given function.

    :param functionClass: Function class for the state (ie, power)
    :param value: Value to set for the function_class
    :param lastUpdateTime: Last time the state was updated (in epoch ms).
    :param functionInstance: Additional information about the function (ie, light-power).
        Default: None
    """

    functionClass: str  # noqa: N815
    value: Any
    lastUpdateTime: int | None = None  # noqa: N815
    functionInstance: str | None = None  # noqa: N815


@dataclass
class AferoCapability:
    """Capability of a given device."""

    functionClass: str  # noqa: N815
    type: str
    schedulable: bool
    functionInstance: str | None = None  # noqa: N815
    _opts: dict[str, Any] = field(default_factory=dict)

    @property
    def options(self) -> dict[str, Any]:
        """Return the options for the capability."""
        return self._opts


@dataclass
class AferoDevice:
    """Mapped Device from an API response."""

    id: str
    device_id: str
    model: str
    device_class: str
    default_name: str
    default_image: str
    friendly_name: str
    capabilities: list[AferoCapability] = field(default_factory=list)
    functions: list[dict] = field(default=list)
    states: list[AferoState] = field(default=list)
    children: list[str] = field(default=list)
    manufacturerName: str | None = field(default=None)  # noqa: N815
    split_identifier: str | None = field(default=None, repr=False)
    version_data: dict[str, str] | None = field(default=None)

    def __hash__(self):
        """Hash."""
        return hash((self.id, self.friendly_name))

    def __post_init__(self):
        """Post init."""
        # Dimmer Switch fix - A switch cannot dim, but a light can
        if self.device_class == "switch" and any(
            state.functionClass == "brightness" for state in self.states
        ):
            self.device_class = "light"
        # Fix exhaust fans
        if (
            self.device_class == "exhaust-fan"
            and self.default_image == "fan-exhaust-icon"
        ):
            self.model = "BF1112"
        # Fix fans
        if self.device_class in ["fan", "ceiling-fan"]:
            if not self.model and self.default_image == "ceiling-fan-snyder-park-icon":
                self.model = "Driskol"
            elif not self.model and self.default_image == "ceiling-fan-vinings-icon":
                self.model = "Vinwood"
            elif (
                self.model == "TBD" and self.default_image == "ceiling-fan-chandra-icon"
            ):
                self.model = "Zandra"
            elif (
                self.model == "TBD"
                and self.default_image == "ceiling-fan-ac-cct-dardanus-icon"
            ):
                self.model = "Nevali"
            elif not self.model and self.default_image == "ceiling-fan-slender-icon":
                self.model = "Tager"
        # Fix lights
        elif self.device_class == "light":
            if self.default_image == "a19-e26-color-cct-60w-smd-frosted-icon":
                self.model = "12A19060WRGBWH2"
            elif self.default_image == "slide-dimmer-icon":
                self.model = "HPDA110NWBP"
            elif self.default_image == "bright-edgelit-flushmount-light-icon":
                self.manufacturerName = "Commercial-Electric"
                self.model = "LCN3002LM-01 WH"
        # Fix switches
        elif self.device_class == "switch":
            if self.default_image == "smart-switch-icon" and self.model == "TBD":
                self.model = "HPSA11CWB"
        # Fix glass doors - Treat as a switch
        elif self.device_class == "glass-door":
            self.device_class = "switch"
            self.manufacturerName = "Feather River Doors"
        # Attempt to fix anything TBD
        if self.model == "TBD" and self.default_name:
            self.model = self.default_name


def transform_capability(capability: dict[str, Any]) -> AferoCapability:
    """Transform a capability dictionary into an AferoCapability."""
    top_level_fields = {
        x.name: capability.get(x.name)
        for x in fields(AferoCapability)
        if not x.name.startswith("_")
    }
    extra_fields = {
        key: capability[key] for key in capability if key not in top_level_fields
    }
    return AferoCapability(**top_level_fields, _opts=extra_fields)


def convert_state(state: dict[str, Any]) -> AferoState:
    """Convert a state dictionary into an AferoState."""
    return AferoState(
        functionClass=state.get("functionClass"),
        value=state.get("value"),
        lastUpdateTime=state.get("lastUpdateTime"),
        functionInstance=state.get("functionInstance"),
    )


def get_afero_device(afero_device: dict[str, Any]) -> AferoDevice:
    """Convert the Afero device definition into a AferoDevice."""
    description = afero_device.get("description", {})
    device = description.get("device", {})
    processed_states: list[AferoState] = []
    processed_states = [
        convert_state(state)
        for state in afero_device.get("state", {}).get("values", [])
    ]
    processed_capabilities = [
        transform_capability(cap) for cap in afero_device.get("capabilities", [])
    ]
    dev_dict = {
        "id": afero_device.get("id"),
        "device_id": afero_device.get("deviceId"),
        "model": device.get("model"),
        "device_class": device.get("deviceClass"),
        "default_name": device.get("defaultName"),
        "default_image": description.get("defaultImage"),
        "friendly_name": afero_device.get("friendlyName"),
        "functions": description.get("functions", []),
        "states": processed_states,
        "capabilities": processed_capabilities,
        "children": afero_device.get("children", []),
        "manufacturerName": device.get("manufacturerName"),
        "version_data": afero_device.get("version_data"),
    }
    return AferoDevice(**dev_dict)


def get_function_from_device(
    functions: list[dict], function_class: str, function_instance: str | None = None
) -> dict | None:
    """Find a function from a device.

    :param functions: List of functions to search through
    :param function_class: Function class to find
    :param function_instance: Function instance to find. Default: None
    """
    for func in functions:
        if func.get("functionClass") != function_class:
            continue
        if func.get("functionInstance") != function_instance:
            continue
        return func
    return None


AferoResource = TypeVar("AferoResource")



"""Test Security System Sensor"""

import copy

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers import event
from aioafero.v1.controllers.security_system import security_system_callback
from aioafero.v1.controllers.security_system_sensor import (
    AferoBinarySensor,
    features,
)
from dataclasses import asdict

from .. import utils

security_system = utils.create_devices_from_data("security-system.json")[1]
security_system_sensors = security_system_callback(
    utils.create_devices_from_data("security-system.json")[1]
).split_devices
security_system_sensor_2 = security_system_sensors[1]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.security_systems_sensors


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data([security_system])
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 3
    dev = mocked_controller["7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2"]
    assert dev.available is True
    assert dev.id == "7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2"
    assert dev.update_id == "7f4e4c01-e799-45c5-9b1a-385433a78edc"
    assert dev.instance == 2
    assert dev.sensors == {}
    assert dev.binary_sensors == {
        "tampered|None": AferoBinarySensor(
            id="tampered|None",
            owner="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
            current_value="Off",
            _error="On",
            unit=None,
            instance=None,
        ),
        "triggered|None": AferoBinarySensor(
            id="triggered|None",
            owner="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
            current_value="On",
            _error="On",
            unit=None,
            instance=None,
        ),
    }
    assert dev.selects == {
        ("bypassType", None): features.SelectFeature(
            selected="Off",
            selects={
                "Off",
                "Manual",
                "On",
            },
            name="Bypass",
        ),
        ("chirpMode", None): features.SelectFeature(
            selected="Off",
            selects={
                "Off",
                "On",
            },
            name="Chime",
        ),
        ("triggerType", None): features.SelectFeature(
            selected="Home/Away",
            selects={
                "Away",
                "Home",
                "Home/Away",
                "Off",
            },
            name="Alarming State",
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller, caplog):
    caplog.set_level("DEBUG")
    await mocked_controller._bridge.events.generate_events_from_data(
        utils.create_hs_raw_from_dump("security-system.json")
    )
    await mocked_controller._bridge.async_block_until_done()
    dev_update = copy.deepcopy(security_system_sensor_2)
    new_states = [
        AferoState(functionClass='chirpMode', value='On', lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='triggerType', value='Away', lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='bypassType', value='Manual', lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='top-level-key', value='security-sensor-config-v2', lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='tampered', value='On', lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='triggered', value='Off', lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='available', value=False, lastUpdateTime=None, functionInstance=None)
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    await mocked_controller._bridge.async_block_until_done()
    assert updates == {
        "select-('bypassType', None)",
        "select-('chirpMode', None)",
        "binary-tampered|None",
        "select-('triggerType', None)",
        "binary-triggered|None",
        "available",
    }
    dev = mocked_controller[security_system_sensor_2.id]
    assert dev.available is False
    assert dev.sensors == {}
    assert dev.binary_sensors == {
        "tampered|None": AferoBinarySensor(
            id="tampered|None",
            owner="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
            current_value="On",
            _error="On",
            unit=None,
            instance=None,
        ),
        "triggered|None": AferoBinarySensor(
            id="triggered|None",
            owner="7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2",
            current_value="Off",
            _error="On",
            unit=None,
            instance=None,
        ),
    }
    assert dev.selects == {
        ("bypassType", None): features.SelectFeature(
            selected="Manual",
            selects={
                "Off",
                "Manual",
                "On",
            },
            name="Bypass",
        ),
        ("chirpMode", None): features.SelectFeature(
            selected="On",
            selects={
                "Off",
                "On",
            },
            name="Chime",
        ),
        ("triggerType", None): features.SelectFeature(
            selected="Away",
            selects={
                "Away",
                "Home",
                "Home/Away",
                "Off",
            },
            name="Alarming State",
        ),
    }


@pytest.mark.asyncio
async def test_update_security_sensor_no_updates(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        utils.create_hs_raw_from_dump("security-system.json")
    )
    await mocked_controller._bridge.async_block_until_done()
    updates = await mocked_controller.update_elem(security_system_sensor_2)
    assert updates == set()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    (
        "device",
        "updates",
        "expected_updates",
    ),
    [
        # Selects are updated
        (
            security_system_sensor_2,
            {
                "selects": {
                    ("chirpMode", None): "On",
                    ("triggerType", None): "Away",
                    ("bypassType", None): "On",
                    ("doesnt_exist", None): "On",
                }
            },
            [
                {
                    "functionClass": "sensor-config",
                    "value": {
                        "security-sensor-config-v2": {
                            "chirpMode": 1,
                            "triggerType": 3,
                            "bypassType": 4,
                        }
                    },
                    "functionInstance": "sensor-2",
                    "lastUpdateTime": 12345,
                }
            ],
        ),
    ],
)
async def test_set_state(device, updates, expected_updates, mocked_controller, mocker):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        utils.create_hs_raw_from_dump("security-system.json")
    )
    await bridge.async_block_until_done()
    # Split devices require a full update during testing
    resp = mocker.AsyncMock()
    dev_update = utils.create_devices_from_data("security-system.json")[1]
    for state in expected_updates:
        utils.modify_state(dev_update, AferoState(**state))
    # Split devices need their IDs correctly set
    json_resp = mocker.AsyncMock()
    json_resp.return_value = {"metadeviceId": security_system.id, "values": [asdict(x) for x in dev_update.states]}
    resp = mocker.AsyncMock()
    resp.json = json_resp
    resp.status = 200
    mocker.patch.object(mocked_controller, "update_afero_api", return_value=resp)
    await mocked_controller.set_state(device.id, **updates)
    await bridge.async_block_until_done()
    dev = mocked_controller["7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2"]
    assert dev.selects[("chirpMode", None)].selected == 'On'
    assert dev.selects[("triggerType", None)].selected == 'Home/Away'
    assert dev.selects[("bypassType", None)].selected == 'On'


@pytest.mark.asyncio
async def test_set_state_bad_device(mocked_controller):
    await mocked_controller.set_state(
        "bad device",
        {
            "selects": {
                ("sensor-2", "chirpMode"): "On",
                ("sensor-2", "triggerType"): "Away",
                ("sensor-2", "bypassType"): "On",
            }
        },
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_set_states_nothing(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        utils.create_hs_raw_from_dump("security-system.json")
    )
    await mocked_controller._bridge.async_block_until_done()
    await mocked_controller.set_state(
        security_system_sensor_2.id,
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_emitting(mocked_bridge):
    # Simulate the discovery process
    await mocked_bridge.generate_devices_from_data([security_system])
    await mocked_bridge.async_block_until_done()
    assert len(mocked_bridge.security_systems_sensors._items) == 3
    dev_update = copy.deepcopy(security_system)
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="sensor-state",
            functionInstance="sensor-2",
            value={
                "security-sensor-state": {
                    "deviceType": 2,
                    "tampered": 0,
                    "triggered": 1,
                    "missing": 1,
                    "versionBuild": 3,
                    "versionMajor": 2,
                    "versionMinor": 0,
                    "batteryLevel": 100,
                }
            },
        ),
    )
    await mocked_bridge.generate_devices_from_data([dev_update])
    await mocked_bridge.async_block_until_done()
    assert len(mocked_bridge.security_systems_sensors._items) == 3
    assert not mocked_bridge.security_systems_sensors._items["7f4e4c01-e799-45c5-9b1a-385433a78edc-sensor-2"].available



"""Base class for Controllers."""

import asyncio
from asyncio.coroutines import iscoroutinefunction
from collections.abc import Callable, Iterator
import contextlib
from dataclasses import dataclass, fields
from datetime import UTC, datetime
import re
import time
from typing import TYPE_CHECKING, Any, Generic, NamedTuple

import aiohttp

from aioafero.device import (
    AferoDevice,
    AferoResource,
    AferoState,
    convert_state,
    get_afero_device,
)
from aioafero.errors import DeviceNotFound, ExceededMaximumRetries
from aioafero.util import process_function
from aioafero.v1 import v1_const
from aioafero.v1.models.features import NumbersFeature, SelectFeature
from aioafero.v1.models.resource import ResourceTypes
from aioafero.v1.models.sensor import AferoBinarySensor, AferoSensor

from .event import AferoEvent, EventCallBackType, EventType

if TYPE_CHECKING:  # pragma: no cover
    from aioafero.v1 import AferoBridgeV1


EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
]

ID_FILTER_ALL = "*"

unit_extractor = re.compile(r"(\d*)(\D*)")


class NumbersName(NamedTuple):
    """Data used for displaying a Number."""

    unit: str
    display_name: str | None = None


class BaseResourcesController(Generic[AferoResource]):
    """Base Controller for Afero IoT Cloud devices."""

    ITEM_TYPE_ID: ResourceTypes | None = None
    ITEM_TYPES: list[ResourceTypes] | None = None
    ITEM_CLS = None
    # functionClass map between controller -> Afero IoT Cloud
    ITEM_MAPPING: dict = {}
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {}
    # Elements that map to numbers. func class / func instance to unit
    ITEM_NUMBERS: dict[tuple[str, str | None], NumbersName] = {}
    # Elements that map to selects func class / func instance to name
    ITEM_SELECTS: dict[tuple[str, str | None], str] = {}
    # Device Split Callbacks
    DEVICE_SPLIT_CALLBACKS: dict[str, callable] = {}

    def __init__(self, bridge: "AferoBridgeV1") -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._items: dict[str, AferoResource] = {}
        self._logger = bridge.logger.getChild(self.ITEM_CLS.__name__)
        self._subscribers: dict[str, list[EventSubscriptionType]] = {ID_FILTER_ALL: []}
        self._initialized: bool = False
        self._item_values = [x.value for x in self.ITEM_TYPES]

    def __getitem__(self, device_id: str) -> AferoResource:
        """Get item by device_id."""
        return self._items[device_id]

    def __iter__(self) -> Iterator[AferoResource]:
        """Iterate items."""
        return iter(self._items.values())

    def __contains__(self, device_id: str) -> bool:
        """Return bool if device_id is in items."""
        return device_id in self._items

    @property
    def items(self) -> list[AferoResource]:
        """Return all items for this resource."""
        return list(self._items.values())

    @property
    def initialized(self) -> bool:
        """Determine if the controller has been initialized."""
        return self._initialized

    @property
    def subscribers(self) -> dict[str, list[EventSubscriptionType]]:
        """Get all subscribers aligned to this controller."""
        return self._subscribers

    async def _handle_event(
        self, evt_type: EventType, evt_data: AferoEvent | None
    ) -> None:
        """Handle incoming event for this resource."""
        if evt_data is None:
            return
        item_id = evt_data.get("device_id", None)
        cur_item = await self._handle_event_type(evt_type, item_id, evt_data)
        if evt_type == EventType.RESOURCE_UPDATE_RESPONSE:
            evt_type = EventType.RESOURCE_UPDATED
        if cur_item:
            await self.emit_to_subscribers(evt_type, item_id, cur_item)

    async def _handle_event_type(
        self, evt_type: EventType, item_id: str, evt_data: AferoEvent
    ) -> AferoResource | list[AferoResource] | None:
        """Determine what to do with the incoming event.

        :param evt_type: Type of event
        :param item_id: ID of the item
        :param evt_data: Event data

        :return: Item after being processed
        """
        if evt_type == EventType.RESOURCE_ADDED:
            self._logger.info(
                "Initializing %s [%s] as a %s",
                evt_data["device"].friendly_name,
                evt_data["device"].id,
                self.ITEM_CLS.__name__,
            )
            cur_item = await self.initialize_elem(evt_data["device"])
            self._items[item_id] = cur_item
            self._bridge.add_device(evt_data["device"].id, self)
        elif evt_type == EventType.RESOURCE_DELETED:
            cur_item = self._items.pop(item_id, evt_data)
            self._bridge.remove_device(evt_data["device_id"])
        elif evt_type in [
            EventType.RESOURCE_UPDATED,
            EventType.RESOURCE_UPDATE_RESPONSE,
        ]:
            # existing item updated
            try:
                cur_item = self.get_device(item_id)
            except DeviceNotFound:
                return None
            if not await self.update_elem(evt_data["device"]) and not evt_data.get(
                "force_forward", False
            ):
                return None
        else:
            # Skip all other events
            return None
        return cur_item

    async def emit_to_subscribers(
        self, evt_type: EventType, item_id: str, item: AferoResource
    ):
        """Emit updates to subscribers.

        :param evt_type: Type of event
        :param item_id: ID of the item
        :param item: Item to emit to subscribers
        """
        subscribers = (
            self._subscribers.get(item_id, []) + self._subscribers[ID_FILTER_ALL]
        )
        for callback, event_filter in subscribers:
            if event_filter is not None and evt_type not in event_filter:
                continue
            # dispatch the full resource object to the callback
            if iscoroutinefunction(callback):
                self._bridge.add_job(asyncio.create_task(callback(evt_type, item)))
            else:
                callback(evt_type, item)

    def get_filtered_devices(self, initial_data: list[dict]) -> list[AferoDevice]:
        """Determine devices that align to the controller."""
        valid_devices: list[AferoDevice] = []
        for element in initial_data:
            if element["typeId"] != self.ITEM_TYPE_ID.value:
                self._logger.debug(
                    "TypeID [%s] does not match %s",
                    element["typeId"],
                    self.ITEM_TYPE_ID.value,
                )
                continue
            device = get_afero_device(element)
            if device.device_class not in self._item_values:
                self._logger.debug(
                    "Device Class [%s] is not contained in %s",
                    device.device_class,
                    self._item_values,
                )
                continue
            valid_devices.append(device)
        return valid_devices

    async def _get_valid_devices(self, initial_data: list[dict]) -> list[AferoDevice]:
        return self.get_filtered_devices(initial_data)

    async def initialize(self) -> None:
        """Initialize controller the controller.

        Initialization process should only occur once. During this process, it will
        subscribe to all updates for the given resources and register any device
        split callbacks for the event controller.
        """
        if self._initialized:
            return
        # subscribe to item updates
        res_filter = tuple(x.value for x in self.ITEM_TYPES)
        self._bridge.events.subscribe(
            self._handle_event,
            resource_filter=res_filter,
        )
        for name, callback in self.DEVICE_SPLIT_CALLBACKS.items():
            self._bridge.events.register_multi_device(name, callback)
        self._initialized = True

    async def initialize_number(
        self, func_def: dict, state: AferoState
    ) -> tuple[tuple[str, str | None], NumbersFeature] | None:
        """Initialize a number from the provided data."""
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_NUMBERS:
            working_def = func_def["values"][0]
            primary_name = self.ITEM_NUMBERS[key].display_name
            if primary_name is None:
                fallback_name = f"{state.functionClass}"
                if state.functionInstance is not None:
                    fallback_name += f"-{state.functionInstance}"
                primary_name = working_def.get("name", fallback_name)
            return key, NumbersFeature(
                value=state.value,
                min=working_def["range"]["min"],
                max=working_def["range"]["max"],
                step=working_def["range"]["step"],
                name=primary_name,
                unit=self.ITEM_NUMBERS[key].unit,
            )
        return None

    async def initialize_select(
        self, functions: list[dict], state: AferoState
    ) -> tuple[tuple[str, str | None], SelectFeature] | None:
        """Initialize a select from the provided data."""
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_SELECTS:
            return key, SelectFeature(
                selected=state.value,
                selects=set(
                    process_function(
                        functions, state.functionClass, state.functionInstance
                    )
                ),
                name=self.ITEM_SELECTS[key],
            )
        return None

    async def initialize_sensor(
        self, state: AferoState, child_id: str
    ) -> AferoSensor | AferoBinarySensor | None:
        """Initialize the sensor.

        :param state: State to update
        :param child_id: device_id of the parent device
        """
        if state.functionClass in self.ITEM_SENSORS:
            value, unit = await self.split_sensor_data(state)
            return AferoSensor(
                id=state.functionClass,
                owner=child_id,
                value=value,
                unit=unit,
            )
        if state.functionClass in self.ITEM_BINARY_SENSORS:
            value, _ = await self.split_sensor_data(state)
            key = f"{state.functionClass}|{state.functionInstance}"
            return AferoBinarySensor(
                id=key,
                owner=child_id,
                instance=state.functionInstance,
                current_value=value,
                _error=self.ITEM_BINARY_SENSORS[state.functionClass],
            )
        return None

    async def update_number(
        self, state: AferoState, cur_item: AferoResource
    ) -> str | None:
        """Update the number if its tracked and a change has been detected.

        :param state: State to update
        :param cur_item: Current item to update
        :return: Identifier of the number that was updated or None
        """
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_NUMBERS:
            if cur_item.numbers[key].value != state.value:
                cur_item.numbers[key].value = state.value
                return f"number-{key}"
        return None

    async def update_select(
        self, state: AferoState, cur_item: AferoResource
    ) -> str | None:
        """Update the select if its tracked and a change has been detected.

        :param state: State to update
        :param cur_item: Current item to update
        :return: Identifier of the select that was updated or None
        """
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_SELECTS:
            if cur_item.selects[key].selected != state.value:
                cur_item.selects[key].selected = state.value
                return f"select-{key}"
        return None

    async def update_sensor(
        self, state: AferoState, cur_item: AferoResource
    ) -> str | None:
        """Update the sensor if its tracked and a change has been detected.

        :param state: State to update
        :param cur_item: Current item to update
        :return: Identifier of the sensor that was updated or None
        """
        if state.functionClass in self.ITEM_SENSORS:
            value, _ = await self.split_sensor_data(state)
            if cur_item.sensors[state.functionClass].value != value:
                cur_item.sensors[state.functionClass].value = value
                return f"sensor-{state.functionClass}"
        elif state.functionClass in self.ITEM_BINARY_SENSORS:
            value, _ = await self.split_sensor_data(state)
            key = f"{state.functionClass}|{state.functionInstance}"
            if cur_item.binary_sensors[key].current_value != value:
                cur_item.binary_sensors[key].current_value = value
                return f"binary-{key}"
        return None

    async def split_sensor_data(self, state: AferoState) -> tuple[Any, str | None]:
        """Split the sensor value and return a tuple of the sensor value and key."""
        if isinstance(state.value, str):
            match = unit_extractor.match(state.value)
            if match and match.group(1) and match.group(2):
                return int(match.group(1)), match.group(2)
        return state.value, self.ITEM_SENSORS.get(state.functionClass, None)

    async def initialize_elem(self, element: AferoDevice) -> None:  # pragma: no cover
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        raise NotImplementedError("Class should implement initialize_elem")

    async def update_elem(self, element: AferoDevice) -> None:  # pragma: no cover
        """Update the Portable AC with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        raise NotImplementedError("Class should implement update_elem")

    def subscribe(
        self,
        callback: EventCallBackType,
        id_filter: str | tuple[str] | None = None,
        event_filter: EventType | tuple[EventType] | None = None,
    ) -> Callable:
        """Subscribe to status changes for this resource type.

        :param callback: callback function to call when an event emits.
        :param id_filter: Optionally provide resource ID(s) to filter events for.
        :param event_filter: Optionally provide EventType(s) as filter.

        Returns:
            function to unsubscribe.

        """
        if not isinstance(event_filter, None | list | tuple):
            event_filter = (event_filter,)

        if id_filter is None:
            id_filter = (ID_FILTER_ALL,)
        elif not isinstance(id_filter, list | tuple):
            id_filter = (id_filter,)

        subscription = (callback, event_filter)

        for id_key in id_filter:
            if id_key not in self._subscribers:
                self._subscribers[id_key] = []
            self._subscribers[id_key].append(subscription)

        # unsubscribe logic
        def unsubscribe():
            for id_key in id_filter:
                if id_key not in self._subscribers:
                    continue
                self._subscribers[id_key].remove(subscription)

        return unsubscribe

    async def _process_state_update(
        self, cur_item: AferoResource, device_id: str, states: list[dict]
    ) -> None:
        dev_states = [
            AferoState(
                functionClass=state["functionClass"],
                value=state["value"],
                functionInstance=state.get("functionInstance"),
                lastUpdateTime=int(datetime.now(UTC).timestamp() * 1000),
            )
            for state in states
        ]
        dummy_update = AferoDevice(
            id=device_id,
            device_id=cur_item.device_information.parent_id,
            model=cur_item.device_information.model,
            device_class=cur_item.device_information.device_class,
            default_image=cur_item.device_information.default_image,
            default_name=cur_item.device_information.default_name,
            friendly_name=cur_item.device_information.name,
            states=dev_states,
        )
        # Update now, but also trigger all chained updates
        await self.update_elem(dummy_update)
        self._bridge.events.add_job(
            AferoEvent(
                type=EventType.RESOURCE_UPDATED,
                device_id=device_id,
                device=dummy_update,
                force_forward=True,
            )
        )

    async def update_afero_api(
        self, device_id: str, states: list[dict]
    ) -> aiohttp.ClientResponse | bool:
        """Update Afero IoT API with the new states.

        :param device_id: Afero IoT Device ID
        :param states: States to manually set

        :return: Response if successful, False otherwise.
        """
        url = self._bridge.generate_api_url(
            v1_const.AFERO_GENERICS["API_DEVICE_STATE_ENDPOINT"].format(
                self._bridge.account_id, str(device_id)
            )
        )
        headers = {
            "host": v1_const.AFERO_CLIENTS[self._bridge.afero_client]["API_DATA_HOST"],
            "content-type": "application/json; charset=utf-8",
        }
        payload = {"metadeviceId": str(device_id), "values": states}
        try:
            res = await self._bridge.request("put", url, json=payload, headers=headers)
        except ExceededMaximumRetries:
            self._logger.warning("Maximum retries exceeded for %s", device_id)
            return False
        else:
            # Bad states provided
            if res.status == 400:
                self._logger.warning(
                    "Invalid update provided for %s using %s", device_id, states
                )
                return False
        return res

    async def update(
        self,
        device_id: str,
        obj_in: AferoResource | None = None,
        states: list[dict] | None = None,
        send_duplicate_states: bool = False,
    ) -> aiohttp.ClientResponse | None:
        """Update Afero IoT with the new data.

        :param device_id: Afero IoT Device ID
        :param obj_in: Afero IoT Resource elements to change
        :param states: States to manually set
        :param send_duplicate_states: Send all states, regardless if there's been a change
        """
        update_id = device_id
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info(
                "Unable to update device %s as it does not exist", device_id
            )
            return None
        # split devices use <elem>.update_id to specify the correct device id
        with contextlib.suppress(AttributeError):
            update_id = cur_item.update_id
        if obj_in:
            device_states = dataclass_to_afero(
                cur_item, obj_in, self.ITEM_MAPPING, send_duplicate_states
            )
            if not device_states:
                self._logger.debug("No states to send. Skipping")
                return None
        else:  # Manually setting states
            device_states = states
        # @TODO - Implement bluetooth logic for update
        if res := await self.update_afero_api(update_id, device_states):
            resp_json = await res.json()
            states = [convert_state(val) for val in resp_json.get("values", [])]
            update_dev = self.generate_update_dev(resp_json["metadeviceId"], states)
            update_dev.id = update_id
            await self._bridge.events.generate_events_from_update(update_dev)
            return res
        return None

    def generate_update_dev(self, device_id: str, states: list[AferoState]) -> dict:
        """Generate update data for the event controller."""
        afero_dev = self._bridge.get_afero_device(device_id)
        afero_dev.states = states
        return afero_dev

    def get_device(self, device_id) -> AferoResource:
        """Lookup the device with the given ID."""
        try:
            return self[device_id]
        except KeyError as err:
            raise DeviceNotFound(device_id) from err


def dataclass_to_afero(
    elem: AferoResource, cls: dataclass, mapping: dict, send_duplicate_states: bool
) -> list[dict]:
    """Convert the current state to be consumed by Afero IoT."""
    states = []
    for f in fields(cls):
        current_feature = getattr(cls, f.name, None)
        if current_feature is None:
            continue
        api_key = mapping.get(f.name, f.name)
        # There is probably a better way to approach this
        field_is_dict = str(f.type).startswith("dict")
        is_tuple_key = False
        if field_is_dict and current_feature and current_feature.keys():
            is_tuple_key = isinstance(list(current_feature.keys())[0], tuple)
        # Tuple keys signify (func_class / func_instance).
        if field_is_dict and is_tuple_key:
            states.extend(
                get_afero_states_from_mapped(
                    elem, f.name, current_feature, send_duplicate_states
                )
            )
        elif field_is_dict and not current_feature:
            continue
        else:
            # We need to determine funcClass / funcInstance when we dump our data
            if (
                current_feature == getattr(elem, f.name, None)
                and not send_duplicate_states
            ):
                continue
            current_feature_value = current_feature
            if hasattr(current_feature, "api_value"):
                current_feature_value = current_feature.api_value
            if not isinstance(current_feature_value, list):
                func_instance = get_afero_instance_for_state(
                    elem, current_feature, api_key
                )
                states.append(
                    get_afero_state_from_feature(
                        api_key, func_instance, current_feature_value
                    )
                )
            else:
                states.extend(get_afero_states_from_list(current_feature_value))
    return states


def get_afero_states_from_mapped(
    element: AferoResource,
    field_name: str,
    update_vals: dict,
    send_duplicate_states: bool,
) -> list[dict]:
    """Convert an update element to dict to be consumed by Afero API."""
    states = []
    current_elems = getattr(element, field_name, None)
    for key, val in update_vals.items():
        if val == current_elems.get(key, None) and not send_duplicate_states:
            continue
        states.append(
            {
                "functionClass": key[0],
                "functionInstance": key[1],
                "lastUpdateTime": int(time.time()),
                "value": val.api_value,
            }
        )
    return states


def get_afero_instance_for_state(
    elem: AferoResource, feature, mapped_afero_key: str | None
) -> str | None:
    """Determine the function instance based on the field data or device."""
    if hasattr(feature, "func_instance") and getattr(feature, "func_instance", None):
        instance = getattr(feature, "func_instance", None)
    elif (
        mapped_afero_key
        and hasattr(elem, "get_instance")
        and elem.get_instance(mapped_afero_key)
    ):
        instance = elem.get_instance(mapped_afero_key)
    else:
        instance = None
    return instance


def get_afero_state_from_feature(
    func_class: str, func_instance: str | None, current_val: Any
) -> dict:
    """Generate a single state from the current data."""
    new_state = {
        "functionClass": func_class,
        "functionInstance": func_instance,
        "lastUpdateTime": int(time.time()),
        "value": None,
    }
    if isinstance(current_val, dict):
        new_state.update(current_val)
    else:
        new_state["value"] = current_val
    return new_state


def get_afero_states_from_list(states: list[dict]) -> list[dict]:
    """Add timestamp to the states.

    Assume the state already has functionClass, functionState, and value
    """
    for state in states:
        state["lastUpdateTime"] = int(time.time())
    return states



import pytest

from aioafero.v1.models import features, DeviceInformation
from aioafero.v1.models.hvac_mixin import HVACMixin


class DummyHVAC(HVACMixin):
    def get_mode_to_check(self) -> str | None:
        return self.hvac_mode.mode

@pytest.fixture
def populatedEntity():
    return DummyHVAC(
        current_temperature=features.CurrentTemperatureFeature(
            temperature=35,
            function_class="temperature",
            function_instance="current-temp",
        ),
        hvac_mode=features.HVACModeFeature(
            mode="cool",
            previous_mode="fan",
            modes={"fan", "auto-cool", "dehumidify", "cool"},
            supported_modes={"fan", "auto-cool", "dehumidify", "cool"},
        ),
        target_temperature_heating=features.TargetTemperatureFeature(
            value=19, step=0.5, min=4, max=32, instance="heating-target"
        ),
        target_temperature_cooling=features.TargetTemperatureFeature(
            value=26, step=0.5, min=10, max=37, instance="cooling-target"
        ),
        target_temperature_auto_heating=features.TargetTemperatureFeature(
            value=18, step=0.5, min=4, max=32, instance="auto-heating-target"
        ),
        target_temperature_auto_cooling=features.TargetTemperatureFeature(
            value=26.5, step=0.5, min=4, max=32, instance="auto-cooling-target"
        ),
        fan_running=False,
        fan_mode=features.ModeFeature(mode="off", modes={"on", "off"}),
        display_celsius=True,
    )


@pytest.mark.parametrize(
    ("mode", "celsius", "expected"), [
        ("cool", True, 26),
        ("heat", True, 19),
        ("dry", True, None),
        ("heat", False, 66),
    ]
)
def test_target_temperature(mode, celsius, expected, populatedEntity, mocker):
    populatedEntity.hvac_mode.mode = mode
    populatedEntity.display_celsius = celsius
    assert populatedEntity.target_temperature == expected


def test_target_temperature_no_feature(populatedEntity):
    populatedEntity.hvac_mode.mode = "cool"
    populatedEntity.target_temperature_cooling = None
    assert populatedEntity.target_temperature is None


@pytest.mark.parametrize(("mode", "expected"), [
    ("cool", features.TargetTemperatureFeature(
            value=26, step=0.5, min=10, max=37, instance="cooling-target"
        )),
    ("heat", features.TargetTemperatureFeature(
            value=19, step=0.5, min=4, max=32, instance="heating-target"
        )),
    ("dry", None),
])
def test__get_target_feature(mode, expected, populatedEntity):
    populatedEntity.hvac_mode.mode = mode
    assert populatedEntity._get_target_feature(mode) == expected

@pytest.mark.parametrize(("mode", "expected", "celsius"), [
    ("cool", 0.5, True),
    ("heat", 0.5, True),
    (None, 0.5, True),
    ("cool", 1, False),
])
def test_target_temperature_step(mode, expected, celsius, populatedEntity):
    populatedEntity.hvac_mode.mode = mode
    populatedEntity.display_celsius = celsius
    assert populatedEntity.target_temperature_step == expected


@pytest.mark.parametrize(("mode", "celsius", "is_set", "expected"), [
    ("cool", True, True, 37),
    ("cool", False, True, 99),
    ("auto", True, True, 32),
    ("cool", True, None, None),
    ("heat", True, True, 32),
])
def test_target_temperature_max(mode, celsius, is_set, expected, populatedEntity):
    populatedEntity.hvac_mode.mode = mode
    populatedEntity.display_celsius = celsius
    if not is_set:
        populatedEntity.target_temperature_cooling = None
        populatedEntity.target_temperature_auto_cooling = None
    assert populatedEntity.target_temperature_max == expected


@pytest.mark.parametrize(("mode", "celsius", "is_set", "expected"), [
    ("heat", True, True, 4),
    ("heat", False, True, 39),
    ("auto", True, True, 4),
    ("heat", True, None, None),
])
def test_target_temperature_min(mode, celsius, is_set, expected, populatedEntity):
    populatedEntity.hvac_mode.mode = mode
    populatedEntity.display_celsius = celsius
    if not is_set:
        populatedEntity.target_temperature_heating = None
        populatedEntity.target_temperature_auto_heating = None
    assert populatedEntity.target_temperature_min == expected


@pytest.mark.parametrize(("current_temperature", "celsius", "expected"), [
    (None, None, None),
    (
        features.CurrentTemperatureFeature(
            temperature=35,
            function_class="temperature",
            function_instance="current-temp",
        ),
        True,
        35,
    ),
    (
        features.CurrentTemperatureFeature(
            temperature=35,
            function_class="temperature",
            function_instance="current-temp",
        ),
        False,
        95,
    ),
])
def test_temperature(populatedEntity, current_temperature, celsius, expected):
    populatedEntity.current_temperature = current_temperature
    populatedEntity.display_celsius = celsius
    assert populatedEntity.temperature == expected



"""aioafero API."""

__all__ = [
    "AferoDevice",
    "AferoError",
    "AferoState",
    "EventType",
    "InvalidAuth",
    "InvalidOTP",
    "InvalidResponse",
    "OTPError",
    "OTPRequired",
    "SecuritySystemError",
    "anonymize_device",
    "anonymize_devices",
    "get_afero_device",
    "v1",
]


from importlib.metadata import PackageNotFoundError, version

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "aioafero"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from . import v1
from .anonomyize_data import anonymize_device, anonymize_devices
from .device import AferoDevice, AferoState, get_afero_device
from .errors import (
    AferoError,
    InvalidAuth,
    InvalidOTP,
    InvalidResponse,
    OTPError,
    OTPRequired,
    SecuritySystemError,
)
from .types import EventType



"""Representation of an Afero Valve and its corresponding updates."""

from dataclasses import dataclass, field

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class Valve(StandardMixin):
    """Representation of an Afero Valve."""

    type: ResourceTypes = ResourceTypes.WATER_TIMER

    open: dict[str | None, features.OpenFeature] = field(default_factory=dict)


@dataclass
class ValvePut:
    """States that can be updated for a Valve."""

    open: features.OpenFeature | None = None



"""Test SecuritySystemController"""

import pytest

from aioafero.errors import SecuritySystemError
from aioafero.device import AferoDevice, AferoState, AferoCapability
from aioafero.v1.controllers import event
from aioafero.v1.controllers.security_system import SecuritySystemController, features, security_system_callback, get_valid_states, get_sensor_ids, get_valid_functions, get_sensor_name, get_model_type

from .. import utils

alarm_panel = utils.create_devices_from_data("security-system.json")[1]


def get_alarm_panel_with_siren() -> AferoDevice:
    alarm_panel_with_siren = utils.create_devices_from_data("security-system.json")[1]
    utils.modify_state(
        alarm_panel_with_siren,
        AferoState(
            functionClass="siren-action",
            functionInstance=None,
            value={"security-siren-action": {"resultCode": 0, "command": 4}},
        ),
    )
    return alarm_panel_with_siren


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    mocker.patch("aioafero.v1.controllers.security_system.UPDATE_TIME", 0)
    return mocked_bridge.security_systems


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(alarm_panel)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "7f4e4c01-e799-45c5-9b1a-385433a78edc"
    assert dev.alarm_state == features.ModeFeature(
        mode="disarmed",
        modes={
            "alarming-sos",
            "learning",
            "arm-started-stay",
            "alarming",
            "arm-stay",
            "disarmed",
            "arm-started-away",
            "triggered",
            "arm-away",
        },
    )
    assert dev.siren_action == features.SecuritySensorSirenFeature(
        result_code=None, command=None
    )
    assert dev.numbers == {
        ("arm-exit-delay", "away"): features.NumbersFeature(
            value=30, min=0, max=300, step=1, name="Arm Exit Delay Away", unit="seconds"
        ),
        ("arm-exit-delay", "stay"): features.NumbersFeature(
            value=0, min=0, max=300, step=1, name="Arm Exit Delay Home", unit="seconds"
        ),
        ("disarm-entry-delay", None): features.NumbersFeature(
            value=60, min=1, max=300, step=1, name="Disarm Entry Delay", unit="seconds"
        ),
        ("siren-alarm-timeout", None): features.NumbersFeature(
            value=180, min=0, max=600, step=30, name="Siren Timeout", unit="seconds"
        ),
        ("temporary-bypass-time", None): features.NumbersFeature(
            value=60, min=10, max=300, step=1, name="Bypass Time", unit="seconds"
        ),
    }
    assert dev.selects == {
        ("bypass-allowed", None): features.SelectFeature(
            selected="now-allowed",
            selects={"now-allowed", "allowed"},
            name="Enable Temporary Bypass",
        ),
        ("song-id", "alarm"): features.SelectFeature(
            selected="preset-01",
            selects={
                "preset-01",
                "preset-02",
                "preset-03",
                "preset-04",
                "preset-05",
                "preset-06",
                "preset-07",
                "preset-08",
                "preset-09",
                "preset-10",
                "preset-11",
                "preset-12",
                "preset-13",
            },
            name="Alarm Noise",
        ),
        ("song-id", "chime"): features.SelectFeature(
            selected="preset-02",
            selects={
                "preset-01",
                "preset-02",
                "preset-03",
                "preset-04",
                "preset-05",
                "preset-06",
                "preset-07",
                "preset-08",
                "preset-09",
                "preset-10",
                "preset-11",
                "preset-12",
                "preset-13",
            },
            name="Chime Noise",
        ),
        ("volume", "chime"): features.SelectFeature(
            selected="volume-01",
            selects={"volume-00", "volume-01", "volume-02", "volume-03", "volume-04"},
            name="Chime Volume",
        ),
        ("volume", "entry-delay"): features.SelectFeature(
            selected="volume-01",
            selects={"volume-00", "volume-01", "volume-02", "volume-03", "volume-04"},
            name="Entry Delay Volume",
        ),
        ("volume", "exit-delay-away"): features.SelectFeature(
            selected="volume-01",
            selects={"volume-00", "volume-01", "volume-02", "volume-03", "volume-04"},
            name="Exit Delay Volume Away",
        ),
        ("volume", "exit-delay-stay"): features.SelectFeature(
            selected="volume-01",
            selects={"volume-00", "volume-01", "volume-02", "volume-03", "volume-04"},
            name="Exit Delay Volume Home",
        ),
        ("volume", "siren"): features.SelectFeature(
            selected="volume-04",
            selects={"volume-00", "volume-01", "volume-02", "volume-03", "volume-04"},
            name="Alarm Volume",
        ),
    }


@pytest.mark.asyncio
async def test_initialize_with_siren(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(get_alarm_panel_with_siren())]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.siren_action == features.SecuritySensorSirenFeature(
        result_code=0,
        command=4,
    )


@pytest.mark.asyncio
async def test_disarm(mocked_controller, mocker):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(get_alarm_panel_with_siren())]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[alarm_panel.id].alarm_state.mode = "arm-away"
    # Setup the response after disarm
    panel = utils.create_devices_from_data("security-system.json")[1]
    new_states = [
        AferoState(
            functionClass="alarm-state", value="disarmed", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(panel, state)
    mocker.patch.object(mocked_controller._bridge, "fetch_device_states", return_value=panel.states)
    # Execute the test
    await mocked_controller.disarm(alarm_panel.id, 1234)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[alarm_panel.id]
    assert dev.alarm_state.mode == "disarmed"
    assert dev.siren_action.result_code is None
    assert dev.siren_action.command is None


@pytest.mark.asyncio
async def test_disarm_invalid_pin(mocked_controller, mocker):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(get_alarm_panel_with_siren())]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[alarm_panel.id].alarm_state.mode = "arm-away"
    # Setup the response after disarm
    panel = utils.create_devices_from_data("security-system.json")[1]
    new_states = [
        AferoState(
            functionClass="alarm-state", value="arm-away", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(panel, state)
    mocker.patch.object(mocked_controller._bridge, "fetch_device_states", return_value=panel.states)
    # Execute the test
    with pytest.raises(SecuritySystemError, match="Disarm PIN was not accepted"):
        await mocked_controller.disarm(alarm_panel.id, 1234)
        await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[alarm_panel.id]
    assert dev.alarm_state.mode == "arm-away"


def enable_sensors(dev):
    for state in dev.states:
        if state.functionClass not in ["sensor-state"] or state.value is None:
            continue
        state.value = {
            'security-sensor-state': {
                'deviceType': 1,
                'tampered': 0,
                'triggered': 0,
                'missing': 0,
                'versionBuild': 3,
                'versionMajor': 2,
                'versionMinor': 0,
                'batteryLevel': 100
            }
        }


def disable_sensors(dev):
    for state in dev.states:
        if state.functionClass not in ["sensor-state"] or state.value is None:
            continue
        state.value = {
            'security-sensor-state': {
                'deviceType': 1,
                'tampered': 1,
                'triggered': 1,
                'missing': 1,
                'versionBuild': 3,
                'versionMajor': 2,
                'versionMinor': 0,
                'batteryLevel': 100
            }
        }

def single_sensor_with_bypass(dev, sensor_id):
    enable_sensors(dev)
    for state in dev.states:
        if state.functionClass == "sensor-state" and state.functionInstance == f"sensor-{sensor_id}":
            state.value = {
                'security-sensor-state': {
                    'deviceType': 1,
                    'tampered': 0,
                    'triggered': 1, # Trigger the device, but it will be bypassed
                    'missing': 0,
                    'versionBuild': 3,
                    'versionMajor': 2,
                    'versionMinor': 0,
                    'batteryLevel': 100
                }
            }
        if state.functionClass == "sensor-config" and state.functionInstance == f"sensor-{sensor_id}":
            state.value = {
                "security-sensor-config-v2": {
                    "chirpMode": 1,
                    "triggerType": 3,
                    "bypassType": 4
                }
            }


def bypass_all_sensors(dev):
    for state in dev.states:
        if state.functionClass == "sensor-state":
            state.value = {
                'security-sensor-state': {
                    'deviceType': 1,
                    'tampered': 0,
                    'triggered': 0,
                    'missing': 0,
                    'versionBuild': 3,
                    'versionMajor': 2,
                    'versionMinor': 0,
                    'batteryLevel': 100
                }
            }
        if state.functionClass == "sensor-config":
            state.value = {
                "security-sensor-config-v2": {
                    "chirpMode": 1,
                    "triggerType": 3,
                    "bypassType": 4
                }
            }


@pytest.mark.asyncio
async def test_arm_home(mocked_controller, mocker):
    dev = utils.create_devices_from_data("security-system.json")[1]
    enable_sensors(dev)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(dev)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    # Setup the response after disarm
    panel = utils.create_devices_from_data("security-system.json")[1]
    new_states = [
        AferoState(
            functionClass="alarm-state", value="arm-started-stay", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(panel, state)
    mocker.patch.object(mocked_controller._bridge, "fetch_device_states", return_value=panel.states)
    # Execute the test
    await mocked_controller.arm_home(alarm_panel.id)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[alarm_panel.id]
    assert dev.alarm_state.mode == "arm-started-stay"


@pytest.mark.asyncio
async def test_arm_away(mocked_controller, mocker):
    dev = utils.create_devices_from_data("security-system.json")[1]
    enable_sensors(dev)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(dev)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    # Setup the response after disarm
    panel = utils.create_devices_from_data("security-system.json")[1]
    new_states = [
        AferoState(
            functionClass="alarm-state", value="arm-started-away", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(panel, state)
    mocker.patch.object(mocked_controller._bridge, "fetch_device_states", return_value=panel.states)
    # Execute the test
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.arm_away(alarm_panel.id)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[alarm_panel.id]
    assert dev.alarm_state.mode == "arm-started-away"


@pytest.mark.asyncio
async def test_arm_bad_sensors(mocked_controller, mocker):
    dev = utils.create_devices_from_data("security-system.json")[1]
    disable_sensors(dev)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(dev)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocker.patch.object(mocked_controller._bridge, "fetch_device_states", return_value=dev.states)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    exp_err = "Sensors are open or unavailable: "
    with pytest.raises(SecuritySystemError, match=exp_err):
        await mocked_controller.arm_away(alarm_panel.id)
        await mocked_controller._bridge.async_block_until_done()


@pytest.mark.asyncio
async def test_alarm_trigger(mocked_controller, mocker):
    dev = utils.create_devices_from_data("security-system.json")[1]
    enable_sensors(dev)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(dev)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    # Setup the response after disarm
    panel = utils.create_devices_from_data("security-system.json")[1]
    new_states = [
        AferoState(
            functionClass="alarm-state", value="alarming-sos", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(panel, state)
    mocker.patch.object(mocked_controller._bridge, "fetch_device_states", return_value=panel.states)
    # Execute the test
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.alarm_trigger(alarm_panel.id)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[alarm_panel.id]
    assert dev.alarm_state.mode == "alarming-sos"


@pytest.mark.asyncio
async def test_empty_update(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(alarm_panel)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    update = utils.create_devices_from_data("security-system.json")[1]
    updates = await mocked_controller.update_elem(update)
    assert updates == set()


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(alarm_panel)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller[alarm_panel.id]
    assert dev.available
    update = utils.create_devices_from_data("security-system.json")[1]
    new_states = [
        AferoState(
            functionClass="alarm-state", value="triggered", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="battery-powered", value="battery-powered", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="arm-exit-delay", value=300, lastUpdateTime=0, functionInstance="away"
        ),
        AferoState(
            functionClass="song-id", value="preset-12", lastUpdateTime=0, functionInstance="alarm"
        ),
        AferoState(
            functionClass="siren-action",
            functionInstance=None,
            value={"security-siren-action": {"resultCode": 0, "command": 4}},
        ),
    ]
    for state in new_states:
        utils.modify_state(update, state)
    updates = await mocked_controller.update_elem(update)
    assert not dev.available
    assert updates == {
        "alarm-state",
        "number-('arm-exit-delay', 'away')",
        "available",
        "select-('song-id', 'alarm')",
        "binary-battery-powered|None",
        "siren-action",
    }
    assert dev.alarm_state.mode == "triggered"
    assert dev.numbers[("arm-exit-delay", "away")].value == 300
    assert dev.selects[("song-id", "alarm")].selected == "preset-12"
    assert dev.binary_sensors["battery-powered|None"].current_value == "battery-powered"
    assert dev.binary_sensors["battery-powered|None"].value is True
    assert dev.siren_action.result_code == 0
    assert dev.siren_action.command == 4


@pytest.mark.asyncio
async def test_update_elem_from_siren(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(get_alarm_panel_with_siren())]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    update = get_alarm_panel_with_siren()
    utils.modify_state(
        update,
        AferoState(
            functionClass="siren-action",
            functionInstance=None,
            value=None,
        ),
    )
    updates = await mocked_controller.update_elem(update)
    assert updates == {"siren-action"}
    dev = mocked_controller[alarm_panel.id]
    assert dev.siren_action == features.SecuritySensorSirenFeature(
        result_code=None, command=None
    )


@pytest.mark.asyncio
async def test_update_elem_from_siren_empty(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(get_alarm_panel_with_siren())]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    update = get_alarm_panel_with_siren()
    updates = await mocked_controller.update_elem(update)
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(alarm_panel)]
    )
    await mocked_controller._bridge.async_block_until_done()
    await mocked_controller.set_state(alarm_panel.id)


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(alarm_panel)]
    )
    await mocked_controller._bridge.async_block_until_done()
    await mocked_controller.set_state(
        alarm_panel.id,
        numbers={("arm-exit-delay", "away"): 300, ("bad", None): False},
        selects={("song-id", "alarm"): "preset-12", ("bad", None): False},
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[alarm_panel.id]
    assert dev.numbers[("arm-exit-delay", "away")].value == 300
    assert dev.selects[("song-id", "alarm")].selected == "preset-12"


@pytest.mark.asyncio
async def test_set_state_bad_device(mocked_controller):
    await mocked_controller.set_state(
        alarm_panel.id,
        numbers={("arm-exit-delay", "away"): 300, ("bad", None): False},
        selects={("song-id", "alarm"): "preset-12", ("bad", None): False},
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_emitting(bridge):
    dev_update = utils.create_devices_from_data("security-system.json")[1]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    await bridge.async_block_until_done()
    assert len(bridge.security_systems._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    await bridge.async_block_until_done()
    assert len(bridge.security_systems._items) == 1
    assert not bridge.security_systems[dev_update.id].available


def test_get_sensor_ids():
    assert get_sensor_ids(alarm_panel) == {1, 2, 4}


def test_get_valid_elements_state():
    assert get_valid_states(alarm_panel.states, 4) == [
        AferoState(functionClass='chirpMode', value="On", lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='triggerType', value="Home/Away", lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='bypassType', value="Off", lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='top-level-key', value="security-sensor-config-v2", lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='battery-level', value=100, lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='tampered', value="Off", lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='triggered', value="Off", lastUpdateTime=None, functionInstance=None),
        AferoState(functionClass='available', value=True, lastUpdateTime=None, functionInstance=None),
    ]


def test_get_valid_functions():
    assert get_valid_functions(alarm_panel.functions, 4) == [
        {'functionClass': 'chirpMode', 'functionInstance': 'sensor-4', 'type': 'category', 'values': [{'name': 'Off'}, {'name': 'On'}]},
        {'functionClass': 'triggerType', 'functionInstance': 'sensor-4', 'type': 'category', 'values': [{'name': 'Off'}, {'name': 'Home'}, {'name': 'Away'}, {'name': 'Home/Away'}]},
        {'functionClass': 'bypassType', 'functionInstance': 'sensor-4', 'type': 'category', 'values': [{'name': 'Off'}, {'name': 'Manual'}, {'name': 'On'}]},
    ]


def test_security_system_callback():
    results = security_system_callback(alarm_panel)
    assert results.remove_original is False
    assert len(results.split_devices) == 3
    assert results.split_devices[0].model == "Security System - Motion Sensor"
    assert results.split_devices[1].model == "Security System - Door/Window Sensor"
    assert results.split_devices[2].model == "Security System - Unknown"


capability_device = AferoDevice(
    id="12345",
    device_id="123456",
    model="its fake!",
    device_class="its fake!",
    default_name="its fake!",
    default_image="its fake!",
    friendly_name="its fake!",
    capabilities=[
        AferoCapability(
            functionClass="sensor-state-for-continue",
            functionInstance="sensor-1",
            type="object",
            schedulable=False,
            _opts={
                "name": "Aaaaa"
            }
        ),
        AferoCapability(
            functionClass="sensor-state",
            functionInstance="sensor-1",
            type="object",
            schedulable=False,
            _opts={
                "name": "Aaaaa"
            }
        )
    ]
)

@pytest.mark.parametrize(
    (("device", "sensor_id", "expected")),
    [
        # No capabilities
        (
            alarm_panel,
            4,
            "Sensor 4"
        ),
        # Capabilities
        (
            capability_device,
            1,
            "Aaaaa",
        )
    ]
)
def test_get_sensor_name(device, sensor_id, expected):
    assert get_sensor_name(device.capabilities, sensor_id) == expected


@pytest.mark.parametrize(
    (("device", "expected")),
    [
        # No valid data
        ([], "Unknown"),
        # Valid data and valid type
        (
            [
                AferoState(
                    functionClass="sensor-state",
                    functionInstance="sensor-4",
                    value={
                        "security-sensor-state": {
                            "deviceType": 1,
                            "tampered": 0,
                            "triggered": 0,
                            "missing": 0,
                            "versionBuild": 3,
                            "versionMajor": 2,
                            "versionMinor": 0,
                            "batteryLevel": 100
                        }
                    },
                )
            ],
            "Motion Sensor",
        ),
        # Valid data and invalid type
        (
            [
                AferoState(
                    functionClass="sensor-state",
                    functionInstance="sensor-4",
                    value={
                        "security-sensor-state": {
                            "deviceType": 3,
                            "tampered": 0,
                            "triggered": 0,
                            "missing": 0,
                            "versionBuild": 3,
                            "versionMajor": 2,
                            "versionMinor": 0,
                            "batteryLevel": 100
                        }
                    },
                )
            ],
            "Unknown",
        ),
    ]
)
def test_get_model_type(device, expected):
    assert get_model_type(device, 4) == expected


@pytest.mark.asyncio
async def test_validate_arm_state(mocked_controller, caplog):
    caplog.set_level("DEBUG")
    bridge = mocked_controller._bridge
    panel = get_alarm_panel_with_siren()
    single_sensor_with_bypass(panel, 4)
    await bridge.events.generate_events_from_data(
        [
            utils.create_hs_raw_from_device(panel)
        ]
    )
    await mocked_controller._bridge.async_block_until_done()
    await mocked_controller.validate_arm_state(panel.id, 4)
    await bridge.async_block_until_done()


@pytest.mark.asyncio
async def test_validate_arm_all_bypassed(mocked_controller, caplog):
    caplog.set_level("DEBUG")
    bridge = mocked_controller._bridge
    panel = get_alarm_panel_with_siren()
    bypass_all_sensors(panel)
    await bridge.events.generate_events_from_data(
        [
            utils.create_hs_raw_from_device(panel)
        ]
    )
    await mocked_controller._bridge.async_block_until_done()
    with pytest.raises(SecuritySystemError, match="No sensors are configured for the requested mode."):
        await mocked_controller.validate_arm_state(panel.id, 4)
        await bridge.async_block_until_done()


@pytest.mark.asyncio
async def test_validate_arm_state_invalid_controller(mocked_controller, mocker):
    bridge = mocked_controller._bridge
    panel = get_alarm_panel_with_siren()
    zandra_light = utils.create_devices_from_data("fan-ZandraFan.json")[1]
    panel.children.append(zandra_light.id)
    await bridge.events.generate_events_from_data(
        [
            utils.create_hs_raw_from_device(zandra_light),
            utils.create_hs_raw_from_device(panel)
        ]
    )
    await mocked_controller._bridge.async_block_until_done()
    exp_err = "Sensors are open or unavailable: "
    with pytest.raises(SecuritySystemError, match=exp_err):
        await mocked_controller.validate_arm_state(panel.id, 4)
        await bridge.async_block_until_done()
